from myhdl import *
from myhdl_lib.utils import assign


def arbiter(rst, clk, req_vec, gnt_vec=None, gnt_idx=None, gnt_vld=None, gnt_rdy=None, ARBITER_TYPE="priority"):
    ''' Wrapper that provides common interface to all arbiters '''
    if ARBITER_TYPE == "priority":
        _arb = arbiter_priority(req_vec, gnt_vec, gnt_idx, gnt_vld)
    elif (ARBITER_TYPE == "roundrobin"):
        _arb = arbiter_roundrobin(rst, clk, req_vec, gnt_vec, gnt_idx, gnt_vld, gnt_rdy)
    else:
        assert "Arbiter: Unknown arbiter type: {}".format(ARBITER_TYPE)

    return _arb


def arbiter_priority(req_vec, gnt_vec=None, gnt_idx=None, gnt_vld=None):
    """ Static priority arbiter: grants the request with highest priority, which is the lower index
            req_vec - (i) vector of request signals, req_vec[0] is with the highest priority
            gnt_vec - (o) optional, vector of grants, one grant per request, only one grant can be active at at time
            gnt_idx - (o) optional, grant index, index of the granted request
            gnt_vld - (o) optional, grant valid, indicate that there is a granted request
    """
    REQ_NUM = len(req_vec)
    gnt_vec_s = Signal(intbv(0)[REQ_NUM:])
    gnt_idx_s = Signal(intbv(0, min=0, max=REQ_NUM))
    gnt_vld_s = Signal(bool(0))

    @always_comb
    def prioroty_encoder():
        gnt_vec_s.next = 0
        gnt_idx_s.next = 0
        gnt_vld_s.next = 0
        for i in range(REQ_NUM):
            if ( req_vec[i]==1 ):
                gnt_vec_s.next[i] = 1
                gnt_idx_s.next = i
                gnt_vld_s.next = 1
                break

    if gnt_vec!=None: _vec = assign(gnt_vec, gnt_vec_s)
    if gnt_idx!=None: _idx = assign(gnt_idx, gnt_idx_s)
    if gnt_vld!=None: _vld = assign(gnt_vld, gnt_vld_s)

    return instances()


def arbiter_roundrobin(rst, clk, req_vec, gnt_vec=None, gnt_idx=None, gnt_vld=None, gnt_rdy=None):
    """ Round Robin arbiter: finds the active request with highest priority and presents its index on the gnt_idx output.
            req_vec - (i) vector of request signals, priority changes dynamically
            gnt_vec - (o) optional, vector of grants, one grant per request, only one grant can be active at at time
            gnt_idx - (o) optional, grant index, index of the granted request
            gnt_vld - (o) optional, grant valid, indicate that there is a granted request
            gnt_rdy - (i) grant ready, indicates that the current grant is consumed and priority should be updated.
                          The priority is updated only if there is a valid grant when gnt_rdy is activated.
                          When priority is updated, the currently granted req_vec[gnt_idx] gets the lowest priority and
                          req_vec[gnt_idx+1] gets the highest priority.
                          gnt_rdy should be activated in the same clock cycle when output the grant is used
    """
    REQ_NUM = len(req_vec)
    ptr = Signal(intbv(0, min=0, max=REQ_NUM))

    gnt_vec_s = Signal(intbv(0)[REQ_NUM:])
    gnt_idx_s = Signal(intbv(0, min=0, max=REQ_NUM))
    gnt_vld_s = Signal(bool(0))

    @always(clk.posedge)
    def ptr_proc():
        if (rst):
            ptr.next = REQ_NUM-1
        elif (gnt_rdy and gnt_vld_s):
            ptr.next = gnt_idx_s

    @always_comb
    def roundrobin_encoder():
        gnt_vec_s.next = 0
        gnt_idx_s.next = 0
        gnt_vld_s.next = 0
        for i in range(REQ_NUM):
            if (i>ptr):
                if ( req_vec[i]==1 ):
                    gnt_vec_s.next[i] = 1
                    gnt_idx_s.next = i
                    gnt_vld_s.next = 1
                    return
        for i in range(REQ_NUM):
            if ( req_vec[i]==1 ):
                gnt_vec_s.next[i] = 1
                gnt_idx_s.next = i
                gnt_vld_s.next = 1
                return

    if gnt_vec!=None: _vec = assign(gnt_vec, gnt_vec_s)
    if gnt_idx!=None: _idx = assign(gnt_idx, gnt_idx_s)
    if gnt_vld!=None: _vld = assign(gnt_vld, gnt_vld_s)

    return instances()


if __name__ == '__main__':
    pass


import unittest

from myhdl import *
from myhdl_lib.stream import checksum
import myhdl_lib.simulation as sim

import random
from math import log, ceil


class TestChecksum(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]


    def testRand(self):
        ''' CHECKSUM: Random '''
        MAX_NUM_BYTES = 200
        BYTES_PER_WORD = [2,4,6,8]
        getDut = sim.DUTer()

        def testbench(BYTES_PER_WORD):
            W16_PER_WORD = BYTES_PER_WORD//2
            SUM_WIDTH = 16 + int(ceil(log(MAX_NUM_BYTES/2,2))) + 1

            clk = sim.Clock(val=0, period=10, units="ns")
            rst = sim.ResetSync(clk=clk, val=0, active=1)

            rx_vld = Signal(bool(0))
            rx_sop = Signal(bool(0))
            rx_eop = Signal(bool(0))
            rx_dat = Signal(intbv(0)[BYTES_PER_WORD*8:])
            rx_mty = Signal(intbv(0, min=0, max=BYTES_PER_WORD))
            sumN = Signal(intbv(0)[SUM_WIDTH:])
            sum16 = Signal(intbv(0)[16:])
            sum16n = Signal(intbv(0)[16:])

            argl = {"rst":rst,
                    "clk":clk,
                    "rx_vld":rx_vld,
                    "rx_sop":rx_sop,
                    "rx_eop":rx_eop,
                    "rx_dat":rx_dat,
                    "rx_mty":rx_mty,
                    "chksum":sum16n,
                    "sum16":sum16,
                    "sumN":sumN,
                    "MAX_BYTES":MAX_NUM_BYTES}

            dut = getDut(checksum, **argl)
            clkgen = clk.gen()

            @instance
            def _stim():
                rx_vld.next = 0
                rx_sop.next = 0
                rx_eop.next = 0
                rx_dat.next = 0
                rx_mty.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                S = intbv(0)[SUM_WIDTH:]
                S16 = S[16:] + S[:16]
                S16n = ~S16 & 0xFFFF
                assert S16n==sum16n, "checksum sum16n: expected {}, detected {}".format(hex(S16n), hex(sum16n))
                assert S16==sum16, "checksum sum16: expected {}, detected {}".format(hex(S16), hex(sum16))
                assert S==sumN, "checksum sumN: expected {}, detected {}".format(hex(S), hex(sumN))

                for _ in range(10):
                    bytes = random.randint(1,MAX_NUM_BYTES)
                    words = int(ceil(float(bytes)/BYTES_PER_WORD))
                    empty = (BYTES_PER_WORD - bytes%BYTES_PER_WORD)%BYTES_PER_WORD
                    S = intbv(0)[SUM_WIDTH:]
                    S16 = S[16:] + S[:16]
                    S16n = ~S16 & 0xFFFF
                    for i in range(words):
                        rx_vld.next = 1
                        rx_sop.next = (i==0)
                        rx_eop.next = (i==words-1)
                        rx_dat.next = random.randint(0,(2**(BYTES_PER_WORD*8))-1)
                        rx_mty.next = empty if i==words-1 else 0
                        d = rx_dat.next
                        if rx_eop.next:
                            d[:] = d & ((intbv(0)[BYTES_PER_WORD*8:].max-1)<<(8*rx_mty.next))
                        # Sum the word
                        for i in range(W16_PER_WORD):
                            S[:] = S[:] + d[16*(i+1):16*i]
                        # Resuce sum to 16 bit
                        S16 = S[:]
                        while S16>S16[16:]:
                            S16[:] = S16[16:] + S16[:16]
                        S16[:] = S16[16:]
                        # Invert
                        S16n = ~S16 & 0xFFFF

                        yield clk.posedge
                        yield delay(1)
                        assert S16n==sum16n, "checksum sum16n: expected {}, detected {}".format(hex(S16n), hex(sum16n))
                        assert S16==sum16, "checksum sum16: expected {}, detected {}".format(hex(S16), hex(sum16))
                        assert S==sumN, "checksum sumN: expected {}, detected {}".format(hex(S), hex(sumN))
                        # Gap
                        rx_vld.next = 0
                        for _ in range(random.randint(0,1)):
                            yield clk.posedge

                yield clk.posedge
                raise StopSimulation
            return instances()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for BPW in BYTES_PER_WORD:
                tb = testbench(BPW)
                Simulation(tb).run()
                del tb


    def testOverflow(self):
        ''' CHECKSUM: Overflow '''
        MAX_NUM_BYTES = 201
        BYTES_PER_WORD = [2,4,6,8]
        getDut = sim.DUTer()

        def testbench(BYTES_PER_WORD):
            W16_PER_WORD = BYTES_PER_WORD//2
            SUM_WIDTH = 16 + int(ceil(log(MAX_NUM_BYTES/2,2))) + 1
            INIT_WIDTH = 16
            SUM_WIDTH += 1

            clk = sim.Clock(val=0, period=10, units="ns")
            rst = sim.ResetSync(clk=clk, val=0, active=1)

            rx_vld = Signal(bool(0))
            rx_sop = Signal(bool(0))
            rx_eop = Signal(bool(0))
            rx_dat = Signal(intbv(0)[BYTES_PER_WORD*8:])
            rx_mty = Signal(intbv(0, min=0, max=BYTES_PER_WORD))
            init_sum = Signal(intbv(0)[INIT_WIDTH:])
            sumN = Signal(intbv(0)[SUM_WIDTH:])
            sum16 = Signal(intbv(0)[16:])
            sum16n = Signal(intbv(0)[16:])

            argl = {"rst":rst,
                    "clk":clk,
                    "rx_vld":rx_vld,
                    "rx_sop":rx_sop,
                    "rx_eop":rx_eop,
                    "rx_dat":rx_dat,
                    "rx_mty":rx_mty,
                    "init_sum":init_sum,
                    "chksum":sum16n,
                    "sum16":sum16,
                    "sumN":sumN,
                    "MAX_BYTES":MAX_NUM_BYTES}

            dut = getDut(checksum, **argl)
            clkgen = clk.gen()

            @instance
            def _stim():
                rx_vld.next = 0
                rx_sop.next = 0
                rx_eop.next = 0
                rx_dat.next = 0
                rx_mty.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                S = intbv(0)[SUM_WIDTH:]
                S16 = S[16:] + S[:16]
                S16n = ~S16 & 0xFFFF
                assert S16n==sum16n, "checksum sum16n: expected {}, detected {}".format(hex(S16n), hex(sum16n))
                assert S16==sum16, "checksum sum16: expected {}, detected {}".format(hex(S16), hex(sum16))
                assert S==sumN, "checksum sumN: expected {}, detected {}".format(hex(S), hex(sumN))

                for k in range(10):
                    bytes = MAX_NUM_BYTES
                    words = int(ceil(float(bytes)/BYTES_PER_WORD))
                    empty = (BYTES_PER_WORD - bytes%BYTES_PER_WORD)%BYTES_PER_WORD
                    init_sum.next = init_sum.max-1-k
                    S = intbv(init_sum.next)[SUM_WIDTH:]
                    S16 = S[16:] + S[:16]
                    S16n = ~S16 & 0xFFFF
                    for i in range(words):
                        rx_vld.next = 1
                        rx_sop.next = (i==0)
                        rx_eop.next = (i==words-1)
                        rx_dat.next = rx_dat.max-1-k
                        rx_mty.next = empty if i==words-1 else 0
                        d = rx_dat.next
                        if rx_eop.next:
                            d = rx_dat.next & ((intbv(0)[BYTES_PER_WORD*8:].max-1)<<(8*rx_mty.next))
                        # Sum the word
                        for i in range(W16_PER_WORD):
                            S[:] = S[:] + d[16*(i+1):16*i]
                        # Resuce sum to 16 bit
                        S16 = S[:]
                        while S16>S16[16:]:
                            S16[:] = S16[16:] + S16[:16]
                        S16[:] = S16[16:]
                        # Invert
                        S16n = ~S16 & 0xFFFF

                        yield clk.posedge
                        yield delay(1)
                        assert S16n==sum16n, "checksum sum16n: expected {}, detected {}".format(hex(S16n), hex(sum16n))
                        assert S16==sum16, "checksum sum16: expected {}, detected {}".format(hex(S16), hex(sum16))
                        assert S==sumN, "checksum sumN: expected {}, detected {}".format(hex(S), hex(sumN))

                        init_sum.next = 0
                        # Gap
                        rx_vld.next = 0
                        for _ in range(random.randint(0,1)):
                            yield clk.posedge

                yield clk.posedge
                raise StopSimulation
            return instances()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for BPW in BYTES_PER_WORD:
                tb = testbench(BPW)
                Simulation(tb).run()
                del tb

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import unittest
import random

from myhdl import *
from myhdl_lib.mem import rom, ram_sp_rf, ram_sp_wf, ram_sp_ar,ram_sdp_rf, ram_sdp_wf, ram_sdp_ar, ram_dp_rf, ram_dp_wf, ram_dp_ar
import myhdl_lib.simulation as sim

def mem_fill(clk, we, addr, di, content):
    for a, d in enumerate(content):
        we.next = 1
        addr.next = a
        di.next = d
        yield clk.posedge
    we.next = 0
    addr.next = 0
    di.next = 0

def mem_verify(clk, addr, do, content, ASYNC_RD=False):
    X = 1 if ASYNC_RD else 2
    for a, d in enumerate(content):
        addr.next = a
        for _ in range(X):
            yield clk.posedge
        assert d==do, "Verify@addr {}: expected={}, detected={} \nDATA:{}".format(a, d, do, content)
    addr.next = 0

def mem_write_and_read(clk, we, addrw, addrr, di, do, content_wr, content_rd, ASYNC_RD=False):
    X = 1 if ASYNC_RD else 2
    for a, d in enumerate(content_wr):
        we.next = 1
        addrw.next = a
        addrr.next = a
        di.next = d
        for _ in range(X):
            yield clk.posedge
        assert content_rd[a]==do, "Write&Read@addr {}: expected={}, detected={} \nDATA:{}".format(a, content_rd[a], do, content_rd)
    we.next = 0
    addrw.next = 0
    addrr.next = 0
    di.next = 0


class TestMem(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testRom(self):
        ''' ROM '''
        DATA_RANGE_MIN = -128
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        addr = Signal(intbv(0, min=0, max=ADDR_MAX))
        dout = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        CONTENT = tuple([random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)])

        def stim():
            @instance
            def _inst():
                yield delay(1)
                addr.next = addr.max-1
                yield delay(10)
                for a in range(ADDR_MAX):
                    addr.next = a
                    yield delay(1)
                    assert CONTENT[a]==dout, "At addr {}: expected {}, detected {}".format(a, CONTENT[a], dout)
                raise StopSimulation
            return _inst


        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            stm = stim()
            dut = getDut(rom, addr=addr, dout=dout, CONTENT=CONTENT)
            Simulation(dut, stm).run()
            del dut, stm


    def testRamSpRf(self):
        ''' RAM: Single-Port, Read-First '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        we = Signal(bool(0))
        addr = Signal(intbv(0, min=0, max=ADDR_MAX))
        di = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        do = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clk = sim.Clock(val=0, period=10, units="ns")
        clkgen = clk.gen()

        def stim():
            @instance
            def _inst():
                yield clk.posedge
                yield mem_fill(clk, we, addr, di, CONTENT[0])
                yield mem_verify(clk, addr, do, CONTENT[0])
                for i in range(1,len(CONTENT)):
                    yield mem_write_and_read(clk, we, addr, addr, di, do, content_wr=CONTENT[i], content_rd=CONTENT[i-1])
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_sp_rf, clk=clk, we=we, addr=addr, di=di, do=do)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del dut, stm


    def testRamSpWf(self):
        ''' RAM: Single-Port, Write-First '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        we = Signal(bool(0))
        addr = Signal(intbv(0, min=0, max=ADDR_MAX))
        di = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        do = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clk = sim.Clock(val=0, period=10, units="ns")
        clkgen = clk.gen()

        def stim():
            @instance
            def _inst():
                yield clk.posedge
                yield mem_fill(clk, we, addr, di, CONTENT[0])
                yield mem_verify(clk, addr, do, CONTENT[0])
                for i in range(1,len(CONTENT)):
                    yield mem_write_and_read(clk, we, addr, addr, di, do, content_wr=CONTENT[i], content_rd=CONTENT[i])
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_sp_wf, clk=clk, we=we, addr=addr, di=di, do=do)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del dut, stm


    def testRamSpAr(self):
        ''' RAM: Single-Port, Asynchronous Read '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        we = Signal(bool(0))
        addr = Signal(intbv(0, min=0, max=ADDR_MAX))
        di = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        do = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clk = sim.Clock(val=0, period=10, units="ns")
        clkgen = clk.gen()

        def stim():
            @instance
            def _inst():
                yield clk.posedge
                yield mem_fill(clk, we, addr, di, CONTENT[0])
                yield mem_verify(clk, addr, do, CONTENT[0], ASYNC_RD=True)
                for i in range(1,len(CONTENT)):
                    yield mem_write_and_read(clk, we, addr, addr, di, do, content_wr=CONTENT[i], content_rd=CONTENT[i-1], ASYNC_RD=True)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_sp_ar, clk=clk, we=we, addr=addr, di=di, do=do)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del dut, stm


    def testRamSdpRf(self):
        ''' RAM: Simple-Dual-Port, Read-First '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        we = Signal(bool(0))
        addrw = Signal(intbv(0, min=0, max=ADDR_MAX))
        addrr = Signal(intbv(0, min=0, max=ADDR_MAX))
        di = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        do = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clk = sim.Clock(val=0, period=10, units="ns")
        clkgen = clk.gen()

        def stim():
            @instance
            def _inst():
                yield clk.posedge
                yield mem_fill(clk, we, addrw, di, CONTENT[0])
                yield mem_verify(clk, addrr, do, CONTENT[0])
                for i in range(1,len(CONTENT)):
                    yield mem_write_and_read(clk, we, addrw, addrr, di, do, content_wr=CONTENT[i], content_rd=CONTENT[i-1])
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_sdp_rf, clk=clk, we=we, addrw=addrw, addrr=addrr, di=di, do=do)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del dut, stm


    def testRamSdpWf(self):
        ''' RAM: Simple-Dual-Port, Write-First '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        we = Signal(bool(0))
        addrw = Signal(intbv(0, min=0, max=ADDR_MAX))
        addrr = Signal(intbv(0, min=0, max=ADDR_MAX))
        di = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        do = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clk = sim.Clock(val=0, period=10, units="ns")
        clkgen = clk.gen()

        def stim():
            @instance
            def _inst():
                yield clk.posedge
                yield mem_fill(clk, we, addrw, di, CONTENT[0])
                yield mem_verify(clk, addrr, do, CONTENT[0])
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clk, we, addrw, addrr, di, do, content_wr=CONTENT[i], content_rd=CONTENT[i])
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_sdp_wf, clk=clk, we=we, addrw=addrw, addrr=addrr, di=di, do=do)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del dut, stm


    def testRamSdpAr(self):
        ''' RAM: Simple-Dual-Port, Asynchronous Read '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        we = Signal(bool(0))
        addrw = Signal(intbv(0, min=0, max=ADDR_MAX))
        addrr = Signal(intbv(0, min=0, max=ADDR_MAX))
        di = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        do = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clk = sim.Clock(val=0, period=10, units="ns")
        clkgen = clk.gen()

        def stim():
            @instance
            def _inst():
                yield clk.posedge
                yield mem_fill(clk, we, addrw, di, CONTENT[0])
                yield mem_verify(clk, addrr, do, CONTENT[0], ASYNC_RD=True)
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clk, we, addrw, addrr, di, do, content_wr=CONTENT[i], content_rd=CONTENT[i-1], ASYNC_RD=True)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_sdp_ar, clk=clk, we=we, addrw=addrw, addrr=addrr, di=di, do=do)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del dut, stm


    def testRamDpRf(self):
        ''' RAM: Dual-Port, Read-First '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        wea = Signal(bool(0))
        web = Signal(bool(0))
        addra = Signal(intbv(0, min=0, max=ADDR_MAX))
        addrb = Signal(intbv(0, min=0, max=ADDR_MAX))
        dia = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        dib = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        doa = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        dob = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clka = sim.Clock(val=0, period=10, units="ns")
        clkb = sim.Clock(val=0, period=17, units="ns")
        clkagen = clka.gen()
        clkbgen = clkb.gen()

        def stim():
            @instance
            def _inst():
                yield clka.posedge
                yield mem_fill(clka, wea, addra, dia, CONTENT[0])
                yield mem_verify(clka, addra, doa, CONTENT[0])
                yield mem_verify(clkb, addrb, dob, CONTENT[0])
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clka, wea, addra, addra, dia, doa, content_wr=CONTENT[i], content_rd=CONTENT[i-1])
                yield mem_fill(clkb, web, addrb, dib, CONTENT[0])
                yield mem_verify(clkb, addrb, dob, CONTENT[0])
                yield mem_verify(clka, addra, doa, CONTENT[0])
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clkb, web, addrb, addrb, dib, dob, content_wr=CONTENT[i], content_rd=CONTENT[i-1])
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_dp_rf, clka=clka, clkb=clkb, wea=wea, web=web, addra=addra, addrb=addrb, dia=dia, dib=dib, doa=doa, dob=dob)
            stm = stim()
            Simulation(clkagen, clkbgen, dut, stm).run()
            del dut, stm


    def testRamDpWf(self):
        ''' RAM: Dual-Port, Write-First '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        wea = Signal(bool(0))
        web = Signal(bool(0))
        addra = Signal(intbv(0, min=0, max=ADDR_MAX))
        addrb = Signal(intbv(0, min=0, max=ADDR_MAX))
        dia = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        dib = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        doa = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        dob = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clka = sim.Clock(val=0, period=10, units="ns")
        clkb = sim.Clock(val=0, period=14, units="ns")
        clkagen = clka.gen()
        clkbgen = clkb.gen()

        def stim():
            @instance
            def _inst():
                yield clka.posedge
                yield mem_fill(clka, wea, addra, dia, CONTENT[0])
                yield mem_verify(clka, addra, doa, CONTENT[0])
                yield mem_verify(clkb, addrb, dob, CONTENT[0])
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clka, wea, addra, addra, dia, doa, content_wr=CONTENT[i], content_rd=CONTENT[i])
                yield mem_fill(clkb, web, addrb, dib, CONTENT[0])
                yield mem_verify(clkb, addrb, dob, CONTENT[0])
                yield mem_verify(clka, addra, doa, CONTENT[0])
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clkb, web, addrb, addrb, dib, dob, content_wr=CONTENT[i], content_rd=CONTENT[i])
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_dp_wf, clka=clka, clkb=clkb, wea=wea, web=web, addra=addra, addrb=addrb, dia=dia, dib=dib, doa=doa, dob=dob)
            stm = stim()
            Simulation(clkagen, clkbgen, dut, stm).run()
            del dut, stm


    def testRamDpAr(self):
        ''' RAM: Dual-Port, Asynchronous Read '''
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 127
        ADDR_MAX = 100

        wea = Signal(bool(0))
        web = Signal(bool(0))
        addra = Signal(intbv(0, min=0, max=ADDR_MAX))
        addrb = Signal(intbv(0, min=0, max=ADDR_MAX))
        dia = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        dib = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        doa = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))
        dob = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX+1))

        CONTENT = [[random.randint(DATA_RANGE_MIN, DATA_RANGE_MAX) for _ in range(ADDR_MAX)] for _ in range(5)]

        clka = sim.Clock(val=0, period=10, units="ns")
        clkb = sim.Clock(val=0, period=14, units="ns")
        clkagen = clka.gen()
        clkbgen = clkb.gen()

        def stim():
            @instance
            def _inst():
                yield clka.posedge
                yield mem_fill(clka, wea, addra, dia, CONTENT[0])
                yield mem_verify(clka, addra, doa, CONTENT[0], ASYNC_RD=True)
                yield mem_verify(clkb, addrb, dob, CONTENT[0], ASYNC_RD=True)
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clka, wea, addra, addra, dia, doa, content_wr=CONTENT[i], content_rd=CONTENT[i-1], ASYNC_RD=True)
                yield mem_fill(clkb, web, addrb, dib, CONTENT[0])
                yield mem_verify(clkb, addrb, dob, CONTENT[0], ASYNC_RD=True)
                yield mem_verify(clka, addra, doa, CONTENT[0], ASYNC_RD=True)
                for i in range(1, len(CONTENT)):
                    yield mem_write_and_read(clkb, web, addrb, addrb, dib, dob, content_wr=CONTENT[i], content_rd=CONTENT[i-1], ASYNC_RD=True)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut = getDut(ram_dp_ar, clka=clka, clkb=clkb, wea=wea, web=web, addra=addra, addrb=addrb, dia=dia, dib=dib, doa=doa, dob=dob)
            stm = stim()
            Simulation(clkagen, clkbgen, dut, stm).run()
            del dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import unittest

from myhdl import *
from myhdl_lib.mux import ls_mux
import myhdl_lib.simulation as sim

import random


class TestLsMux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testLsMux2(self):
        ''' LS_MUX: 5 inputs'''
        DMAX = 100
        NUM_INPUTS = 5

        def ls_mux_top(sel, 
                       di_00, di_01, di_02,
                       di_10, di_11, di_12,
                       di_20, di_21, di_22,
                       di_30, di_31, di_32,
                       di_40, di_41, di_42,
                       do_0, do_1, do_2):
            ''' Needed when ls_mux is co-simulated as top level'''
            lsls_di_00 = Signal(bool(0))
            lsls_di_01 = Signal(intbv(0,min=0,max=DMAX))
            lsls_di_02 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_di_10 = Signal(bool(0))
            lsls_di_11 = Signal(intbv(0,min=0,max=DMAX))
            lsls_di_12 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_di_20 = Signal(bool(0))
            lsls_di_21 = Signal(intbv(0,min=0,max=DMAX))
            lsls_di_22 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_di_30 = Signal(bool(0))
            lsls_di_31 = Signal(intbv(0,min=0,max=DMAX))
            lsls_di_32 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_di_40 = Signal(bool(0))
            lsls_di_41 = Signal(intbv(0,min=0,max=DMAX))
            lsls_di_42 = Signal(intbv(0,min=-DMAX,max=DMAX))

            ls_do_0 = Signal(bool(0))
            ls_do_1 = Signal(intbv(0,min=0,max=DMAX))
            ls_do_2 = Signal(intbv(0,min=-DMAX,max=DMAX))

            @always_comb
            def _assign():
                lsls_di_00.next = di_00
                lsls_di_01.next = di_01
                lsls_di_02.next = di_02
                lsls_di_10.next = di_10
                lsls_di_11.next = di_11
                lsls_di_12.next = di_12
                lsls_di_20.next = di_20
                lsls_di_21.next = di_21
                lsls_di_22.next = di_22
                lsls_di_30.next = di_30
                lsls_di_31.next = di_31
                lsls_di_32.next = di_32
                lsls_di_40.next = di_40
                lsls_di_41.next = di_41
                lsls_di_42.next = di_42

                do_0.next = ls_do_0
                do_1.next = ls_do_1
                do_2.next = ls_do_2

            lsls_di = [[lsls_di_00, lsls_di_01, lsls_di_02],
                       [lsls_di_10, lsls_di_11, lsls_di_12],
                       [lsls_di_20, lsls_di_21, lsls_di_22],
                       [lsls_di_30, lsls_di_31, lsls_di_32],
                       [lsls_di_40, lsls_di_41, lsls_di_42]]

            ls_do = [ls_do_0, ls_do_1, ls_do_2]

            _ls_mux = ls_mux(sel=sel, lsls_di=lsls_di, ls_do=ls_do)
            return instances()


        sel = Signal(intbv(0,min=0,max=NUM_INPUTS))
        lsls_di = [[Signal(bool(0)),
                    Signal(intbv(0,min=0,max=DMAX)),
                    Signal(intbv(0,min=-DMAX,max=DMAX))] for _ in range(NUM_INPUTS)]

        ls_do = [Signal(bool(0)),
                 Signal(intbv(0,min=0,max=DMAX)),
                 Signal(intbv(0,min=-DMAX,max=DMAX))]

        argl = {"sel":sel, 
                "di_00":lsls_di[0][0], "di_01":lsls_di[0][1], "di_02":lsls_di[0][2],
                "di_10":lsls_di[1][0], "di_11":lsls_di[1][1], "di_12":lsls_di[1][2],
                "di_20":lsls_di[2][0], "di_21":lsls_di[2][1], "di_22":lsls_di[2][2],
                "di_30":lsls_di[3][0], "di_31":lsls_di[3][1], "di_32":lsls_di[3][2],
                "di_40":lsls_di[4][0], "di_41":lsls_di[4][1], "di_42":lsls_di[4][2],
                "do_0":ls_do[0],"do_1":ls_do[1],"do_2":ls_do[2]}

        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(NUM_INPUTS):
                    for j in range(3):
                        lsls_di[i][j].next = 0
                yield delay(10)

                for _ in range(2):
                    d = [[random.randint(0,1),
                          random.randint(0,DMAX-1),
                          random.randint(-DMAX,DMAX-1)] for _ in range(NUM_INPUTS)]
                    for i in range(NUM_INPUTS):
                        for j in range(3):
                            lsls_di[i][j].next = d[i][j]

                    for s in range(NUM_INPUTS):
                        sel.next = s
                        yield delay(10)
                        assert d[s][0] == ls_do[0], "Mux output (sel={}): expected {}, detected {}".format(sel, d[s], [int(x) for x in ls_do])
                        assert d[s][1] == ls_do[1], "Mux output (sel={}): expected {}, detected {}".format(sel, d[s], [int(x) for x in ls_do])
                        assert d[s][2] == ls_do[2], "Mux output (sel={}): expected {}, detected {}".format(sel, d[s], [int(x) for x in ls_do])

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(ls_mux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


from _DUTer import DUTer
from clock import Clock
from reset import ResetSync
from payload_generator import payload_generator

__all__ =["DUTer",
          "Clock",
          "ResetSync",
          "payload_generator"]


from myhdl import SignalType, instance, delay


# This code is a shameless copy of Christopher Felton's _clock class: https://github.com/cfelton/minnesota/blob/master/mn/system/_clock.py
class Clock(SignalType): 
    ''' Clock signal equipped with clock generator '''

    _frequency_units = {"Hz":1, "kHz":1e3, "Mhz":1e6, "GHz":1e9}
    _periode_units = {"sec":1,"s":1, "ms":1e-3, "us":1e-6, "ns":1e-9}

    def __init__(self, val, frequency=None, period=10, units="ns", ratio_high_to_low=0.5):
        ''' Sets clock parameters
                val - initial clock value
                frequency - clock frequency
                period - clock period, used if clock frequency is not defined
                units - units in which frequency or period is defined
                        allowed frequency units: "Hz", "kHz", "Mhz", "GHz"
                        allowed period units: "sec","s", "ms":, "us", "ns"
                ratio_high_to_low - clock signal ration, 0 < ratio_high_to_low < 1
        '''
        if frequency is not None:
            assert Clock._frequency_units.has_key(units), "Unknown frequency units {}. Allowed values: {}".format(units, Clock._frequency_units.keys())
            self._frequency_Hz = frequency*Clock._frequency_units[units]
            self._period_sec = 1.0/frequency
        else:
            assert Clock._periode_units.has_key(units), "Unknown time units {}. Allowed values: {}".format(units, Clock._periode_units.keys())
            self._period_sec = period*Clock._periode_units[units]
            self._frequency_Hz = 1.0/period

        self._periode_ns = self._period_sec*1e9

        assert (0 < ratio_high_to_low), "Expected 0 < ratio_high_to_low, got ratio_high_to_low={}".format(ratio_high_to_low)
        assert (ratio_high_to_low < 1), "Expected ratio_high_to_low < 1, got ratio_high_to_low={}".format(ratio_high_to_low)


        self._periode_high_ns = long(self._periode_ns * ratio_high_to_low)
        self._periode_low_ns = long(self._periode_ns - self._periode_high_ns)

        self._periode_ns = self._periode_high_ns + self._periode_low_ns
        self.periode_sec = self._periode_ns*1e-9
        self._frequency_Hz = 1e9/self._periode_ns

        SignalType.__init__(self, bool(val))

    @property
    def frequency(self):
        return self._frequency_Hz

    @property
    def period(self):
        return self._period_sec


    def gen(self):
        @instance
        def clkgen():
            if self.val:
                self.next = True
                yield delay(self._periode_high_ns)
            while True:
                self.next = False
                yield delay(self._periode_low_ns)
                self.next = True
                yield delay(self._periode_high_ns)

        return clkgen


if __name__ == '__main__':
    pass


from myhdl import *
from mem   import ram_sdp_ar



def fifo(rst, clk, full, we, din, empty, re, dout, afull=None, aempty=None, afull_th=None, aempty_th=None, ovf=None, udf=None, count=None, count_max=None, depth=None, width=None):
    """ Synchronous FIFO

        Input  interface: full,  we, din
        Output interface: empty, re, dout
            It s possible to set din and dout to None. Then the fifo width will be 0 and the fifo will contain no storage.

        Extra interface:
        afull     (o) - almost full flag, asserted when the number of empty cells <= afull_th
        aempty    (o) - almost empty flag, asserted when the number of full cells <= aempty_th

        afull_th  (i) - almost full threshold, in terms of fifo cells; signal or constant; Optional, default depth/2
        aempty_th (i) - almost empty threshold, in terms of fifo cells; signal or constant; Optional, default depth/2

        count     (o) - number of occupied fifo cells

        count_max (o) - max number of occupied fifo cells reached since the last reset
        ovf       (o) - overflow flag, set at the first write in a full fifo, cleared at reset
        udf       (o) - underflow flag, set at the first read from an empty fifo, cleared at reset

        Parameters:
        depth         - fifo depth, must be >= 1; if not set or set to `None` default value 2 is used
        width         - data width in bits, must be >= 0; if not set or set to `None` the `din` width is used

    """

    if (width == None):
        width = 0
        if din is not None:
            width = len(din)
    if (depth == None):
        depth = 2

    full_flg        = Signal(bool(1))
    empty_flg       = Signal(bool(1))
    we_safe         = Signal(bool(0))
    re_safe         = Signal(bool(0))

    rd_ptr          = Signal(intbv(0, min=0, max=depth))
    rd_ptr_new      = Signal(intbv(0, min=0, max=depth))
    wr_ptr          = Signal(intbv(0, min=0, max=depth))
    wr_ptr_new      = Signal(intbv(0, min=0, max=depth))

    @always_comb
    def safe_read_write():
        full.next       = full_flg
        empty.next      = empty_flg
        we_safe.next    = we and not full_flg
        re_safe.next    = re and not empty_flg


    #===========================================================================
    # Write, Read, Full, Empty
    #===========================================================================
    @always_comb
    def ptrs_new():
        rd_ptr_new.next     = ((rd_ptr + 1) % depth)
        wr_ptr_new.next     = ((wr_ptr + 1) % depth)

    @always(clk.posedge)
    def state_main():
        if (rst):
            wr_ptr.next     = 0
            rd_ptr.next     = 0
            full_flg.next   = 0
            empty_flg.next  = 1
        else:
            # Write pointer
            if (we_safe): wr_ptr.next = wr_ptr_new
            # Read pointer
            if (re_safe): rd_ptr.next = rd_ptr_new
            # Empty flag
            if (we_safe):
                empty_flg.next  = 0
            elif (re_safe and (rd_ptr_new == wr_ptr)):
                empty_flg.next  = 1
            # Full flag
            if (re_safe):
                full_flg.next   = 0
            elif (we_safe and (wr_ptr_new == rd_ptr)):
                full_flg.next   = 1


    #===========================================================================
    # Count, CountMax
    #===========================================================================
    ''' Count '''
    if (count != None) or (count_max != None) or (afull != None) or (aempty != None):
        count_r = Signal(intbv(0, min=0, max=depth+1))
        count_new  = Signal(intbv(0, min=-1, max=depth+2))

        if (count != None):
            assert count.max > depth
            @always_comb
            def count_out():
                count.next = count_r

        @always_comb
        def count_comb():
            if   (we_safe and not re_safe):
                    count_new.next = count_r + 1
            elif (not we_safe and re_safe):
                    count_new.next = count_r - 1
            else:
                    count_new.next = count_r

        @always(clk.posedge)
        def count_proc():
            if (rst):
                count_r.next = 0
            else:
                count_r.next = count_new

    ''' Count max '''
    if (count_max != None):
        assert count_max.max > depth
        count_max_r = Signal(intbv(0, min=0,max=count_max.max))
        @always(clk.posedge)
        def count_max_proc():
            if (rst):
                count_max_r.next = 0
            else:
                if (count_max_r < count_new):
                    count_max_r.next = count_new

        @always_comb
        def count_max_out():
            count_max.next = count_max_r

    #===========================================================================
    # AlmostFull, AlmostEmpty
    #===========================================================================
    ''' AlmostFull flag '''
    if (afull != None):
        if (afull_th == None):
            afull_th = depth//2
        @always(clk.posedge)
        def afull_proc():
            if (rst):
                afull.next = 0
            else:
                afull.next = (count_new >= depth-afull_th)

    ''' AlmostEmpty flag '''
    if (aempty != None):
        if (aempty_th == None):
            aempty_th = depth//2
        @always(clk.posedge)
        def aempty_proc():
            if (rst):
                aempty.next = 1
            else:
                aempty.next = (count_new <=  aempty_th)


    #===========================================================================
    # Overflow, Underflow
    #===========================================================================
    ''' Overflow flag '''
    if (ovf != None):
        @always(clk.posedge)
        def ovf_proc():
            if (rst):
                ovf.next = 0
            else:
                if (we and full_flg ):
                    ovf.next = 1

    ''' Underflow flag '''
    if (udf != None):
        @always(clk.posedge)
        def udf_proc():
            if (rst):
                udf.next = 0
            else:
                if (re and empty_flg):
                    udf.next = 1

    if width>0:
        #===========================================================================
        # Memory instance
        #===========================================================================
        mem_we      = Signal(bool(0))
        mem_addrw   = Signal(intbv(0, min=0, max=depth))
        mem_addrr   = Signal(intbv(0, min=0, max=depth))
        mem_di      = Signal(intbv(0)[width:0])
        mem_do      = Signal(intbv(0)[width:0])

        # RAM: Simple-Dual-Port, Asynchronous read
        mem = ram_sdp_ar(   clk     = clk,
                            we      = mem_we,
                            addrw   = mem_addrw,
                            addrr   = mem_addrr,
                            di      = mem_di,
                            do      = mem_do )

        @always_comb
        def mem_connect():
            mem_we.next         = we_safe
            mem_addrw.next      = wr_ptr
            mem_addrr.next      = rd_ptr
            mem_di.next         = din
            dout.next           = mem_do

    return instances()



def convert(depth=256, width=8):
    rst, clk   = [Signal(bool(0)) for _ in range(2)]
    din, dout  = [Signal(intbv(0)[width:]) for _ in range(2)]

    re, we, empty, full, afull, aempty = [Signal(bool(0)) for _ in range(6)]
    count       = Signal(intbv(0, min=0, max=depth+1))
    count_max   = Signal(intbv(0, min=0, max=depth+1))
    afull_th    = Signal(intbv(0, min=0, max=depth+1))
    aempty_th   = Signal(intbv(0, min=0, max=depth+1))
    ovf         = Signal(bool(0))
    udf         = Signal(bool(0))

    toVerilog(fifo, rst, clk, we, din, full, re, dout, empty, afull, aempty, count, afull_th, aempty_th, ovf, udf, count_max, depth=depth, width=width)



if __name__ == "__main__":
    convert()



from myhdl import *
from myhdl_lib.arbiter import arbiter
from myhdl_lib.utils import assign

'''
    Handshake

     -------      ready      -------
    |       |<--------------|       |
    |       |               |       |
    |  src  |     valid     |  dst  |
    |       |-------------->|       |
    |       |               |       |
    |       |     data      |       |
    |       |==============>|       |
     -------                 -------

              ___   ___   ___   ___   ___   ___   ___   ___   ___   ___   ___   ___   ___   ___   ___   
    clk   ___|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |__|  |___
              ______________________________            ____________      ______      ______            
  ready   ___|                             |___________|           |_____|     |_____|     |____________
                    ____________      ______      ______________________________            ______      
  valid   _________|           |_____|     |_____|                             |___________|     |______
                    _____ _____       _____       ___________ _____ ___________             _____        
   data   xxxxxxxxxX_____X_____XxxxxxX_____XxxxxxX___________X_____X___________XxxxxxxxxxxxX_____Xxxxxxxx
                      ^     ^           ^                 ^     ^           ^                           
  transactions        ^     ^           ^                 ^     ^           ^                           


    Simple interface used to synchronize interactions/data transfers between modules. 
    Signals:
        ready (or rdy) - indicates that in the current clock cycle the destination module is ready to receive
        valid (or vld) - indicates that in the current clock cycle the source module is ready to transmit (has valid data)
        data (or dat) - data to be transmitted

    - Signals "valid" and "ready" advertise *state* (as opposite to notifying about event)

    - The names "valid" and "ready" imply notion of direction by suggesting source and destination.
      Such notion is natural and intuitive when we have data transfer in one direction, which is 
      often the case. Then we can clearly identify data source and data destination, the source advertises 
      the validity of its data via the "valid" signal, and the destination advertises its readiness 
      to accept data via its "ready" signal.

      But unidirectional data transfer is not the only case when we need synchronization between modules. 
      Synchronization may be needed when there is no data transfer (just synchronization), or synchronization 
      may be needed when data is transferred in both directions (modules exchange data). In these two cases 
      the notion of source and destination is lost and the signal names "valid" and "ready" are not natural 
      and intuitive. Without the notion of direction both modules that need to synchronize their operation 
      become equal, and they drive the "ready" and "valid" signals to advertise their readiness. In that 
      sense both signals have meaning of "I_am_ready". From a point of view of a module, the signal driven 
      by the module has meaning of "I_am_ready" and the signal received from the other module has meaning of 
      "The_other_is_ready". Following that notion we can give the signals more intuitive local names like 
      "local_ready" and "foreign_ready", but this will solve the problem only on local level, level of a module,
      since the notion of local and foreign are relative - what is local for one module is foreign for the other 
      and vice versa. We still do not have unique, design independent names for the signals between the modules.

      For the moment we use the ready/valid naming, until a better one is proposed.


'''


def hs_join(ls_hsi, hso):
    """ [Many-to-one] Synchronizes (joins) a list of input handshake interfaces: output is ready when ALL inputs are ready
            ls_hsi - (i) list of input handshake tuples (ready, valid)
            hso    - (o) an output handshake tuple (ready, valid)
    """
    N = len(ls_hsi)
    ls_hsi_rdy, ls_hsi_vld = zip(*ls_hsi)
    ls_hsi_rdy, ls_hsi_vld = list(ls_hsi_rdy), list(ls_hsi_vld)
    hso_rdy, hso_vld = hso

    @always_comb
    def _hsjoin():
        all_vld = True
        for i in range(N):
            all_vld = all_vld and ls_hsi_vld[i]
        hso_vld.next = all_vld
        for i in range(N):
            ls_hsi_rdy[i].next = all_vld and hso_rdy

    return _hsjoin


def hs_fork(hsi, ls_hso):
    """ [One-to-many] Synchronizes (forks) to a list of output handshake interfaces: input is ready when ALL outputs are ready
            hsi    - (i) input handshake tuple (ready, valid)
            ls_hso - (o) list of output handshake tuples (ready, valid)
    """
    N = len(ls_hso)
    hsi_rdy, hsi_vld = hsi
    ls_hso_rdy, ls_hso_vld = zip(*ls_hso)
    ls_hso_rdy, ls_hso_vld = list(ls_hso_rdy), list(ls_hso_vld)

    @always_comb
    def _hsfork():
        all_rdy = True
        for i in range(N):
            all_rdy = all_rdy and ls_hso_rdy[i]
        hsi_rdy.next = all_rdy
        for i in range(N):
            ls_hso_vld[i].next = all_rdy and hsi_vld

    return _hsfork


def hs_mux(sel, ls_hsi, hso):
    """ [Many-to-one] Multiplexes a list of input handshake interfaces
            sel    - (i) selects an input handshake interface to be connected to the output
            ls_hsi - (i) list of input handshake tuples (ready, valid)
            hso    - (o) output handshake tuple (ready, valid)
    """
    N = len(ls_hsi)
    ls_hsi_rdy, ls_hsi_vld = zip(*ls_hsi)
    ls_hsi_rdy, ls_hsi_vld = list(ls_hsi_rdy), list(ls_hsi_vld)
    hso_rdy, hso_vld = hso

    @always_comb
    def _hsmux():
        hso_vld.next = 0
        for i in range(N):
            ls_hsi_rdy[i].next = 0
            if i == sel:
                hso_vld.next = ls_hsi_vld[i]
                ls_hsi_rdy[i].next = hso_rdy

    return _hsmux


def hs_demux(sel, hsi, ls_hso):
    """ [One-to-many] Demultiplexes to a list of output handshake interfaces
            sel    - (i) selects an output handshake interface to connect to the input
            hsi    - (i) input handshake tuple (ready, valid)
            ls_hso - (o) list of output handshake tuples (ready, valid)
    """
    N = len(ls_hso)
    hsi_rdy, hsi_vld = hsi
    ls_hso_rdy, ls_hso_vld = zip(*ls_hso)
    ls_hso_rdy, ls_hso_vld = list(ls_hso_rdy), list(ls_hso_vld)

    @always_comb
    def _hsdemux():
        hsi_rdy.next = 0
        for i in range(N):
            ls_hso_vld[i].next = 0
            if i == sel:
                hsi_rdy.next = ls_hso_rdy[i]
                ls_hso_vld[i].next = hsi_vld

    return _hsdemux


def hs_arbmux(rst, clk, ls_hsi, hso, sel, ARBITER_TYPE="priority"):
    """ [Many-to-one] Arbitrates a list of input handshake interfaces.
        Selects one of the active input interfaces and connects it to the output.
        Active input is an input interface with asserted "valid" signal
            ls_hsi - (i) list of input handshake tuples (ready, valid)
            hso    - (o) output handshake tuple (ready, valid)
            sel    - (o) indicates the currently selected input handshake interface
            ARBITER_TYPE - selects the arbiter type to be used, "priority" or "roundrobin"
    """
    N = len(ls_hsi)
    ls_hsi_rdy, ls_hsi_vld = zip(*ls_hsi)
    ls_hsi_vld = list(ls_hsi_vld)

    # Needed to avoid: "myhdl.ConversionError: Signal in multiple list is not supported:"
    ls_vld = [Signal(bool(0)) for _ in range(N)]
    _a = [assign(ls_vld[i], ls_hsi_vld[i]) for i in range(N)]

    sel_s = Signal(intbv(0, min=0, max=N))
    @always_comb
    def _sel():
        sel.next = sel_s

    priority_update = None

    if (ARBITER_TYPE == "roundrobin"):
        hso_rdy, hso_vld = hso
        priority_update = Signal(bool(0))

        @always_comb
        def _prio():
            priority_update.next = hso_rdy and hso_vld

    _arb = arbiter(rst=rst, clk=clk, req_vec=ls_vld, gnt_idx=sel_s, gnt_rdy=priority_update, ARBITER_TYPE=ARBITER_TYPE)

    _mux = hs_mux(sel=sel_s, ls_hsi=ls_hsi, hso=hso)

    return instances()


def hs_arbdemux(rst, clk, hsi, ls_hso, sel, ARBITER_TYPE="priority"):
    """ [One-to-many] Arbitrates a list output handshake interfaces
        Selects one of the active output interfaces and connects it to the input.
        Active is an output interface with asserted "ready" signal
            hsi    - (i) input handshake tuple (ready, valid)
            ls_hso - (o) list of output handshake tuples (ready, valid)
            sel    - (o) indicates the currently selected output handshake interface
            ARBITER_TYPE - selects the type of arbiter to be used, "priority" or "roundrobin"
    """
    N = len(ls_hso)
    ls_hso_rdy, ls_hso_vld = zip(*ls_hso)
    ls_hso_rdy = list(ls_hso_rdy)

    # Needed to avoid: "myhdl.ConversionError: Signal in multiple list is not supported:"
    ls_rdy = [Signal(bool(0)) for _ in range(N)]
    _a = [assign(ls_rdy[i], ls_hso_rdy[i]) for i in range(N)]

    sel_s = Signal(intbv(0, min=0, max=len(ls_rdy)))
    @always_comb
    def _sel():
        sel.next = sel_s

    priority_update = None
    if (ARBITER_TYPE == "roundrobin"):
        shi_rdy, hsi_vld = hsi
        priority_update = Signal(bool(0))

        @always_comb
        def _prio():
            priority_update.next = shi_rdy and hsi_vld

    _arb = arbiter(rst=rst, clk=clk, req_vec=ls_rdy, gnt_idx=sel_s, gnt_rdy=priority_update, ARBITER_TYPE=ARBITER_TYPE)

    _demux = hs_demux(sel_s, hsi, ls_hso)

    return instances()





if __name__ == '__main__':
    pass


import unittest

from myhdl import *
from myhdl_lib.arbiter import arbiter_priority
import myhdl_lib.simulation as sim


class TestArbiterPriority(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]


    def testOneHot(self):
        ''' ARBITER_RPIORITY: One Hot requests'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                req_vec.next = 0
                yield delay(10)
                check(vec=0, idx=0, vld=0)

                for i in range(NUM_REQ):
                    req_vec.next = 1 << i
                    yield delay(10)
                    check(vec=1<<i, idx=i, vld=1)

                for i in range(NUM_REQ-1,-1,-1):
                    req_vec.next = 1 << i
                    yield delay(10)
                    check(vec=1<<i, idx=i, vld=1)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut = getDut(arbiter_priority, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld)
            stm = stim()
            Simulation(dut, stm).run()
            del dut, stm


    def testJohnson(self):
        ''' ARBITER_RPIORITY: Johnson requests'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                req_vec.next = 0
                yield delay(10)
                check(vec=0, idx=0, vld=0)

                mask = req_vec.max-1
                for i in range(NUM_REQ):
                    req_vec.next = (mask << i) & mask
                    yield delay(10)
                    check(vec=1<<i, idx=i, vld=1)

                for i in range(NUM_REQ-1,-1,-1):
                    req_vec.next = mask >> i
                    yield delay(10)
                    check(vec=1<<0, idx=0, vld=1)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut = getDut(arbiter_priority, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld)
            stm = stim()
            Simulation(dut, stm).run()
            del dut, stm


    def testCount(self):
        ''' ARBITER_RPIORITY: Count'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                req_vec.next = 0
                yield delay(10)
                check(vec=0, idx=0, vld=0)

                x = intbv(0)[NUM_REQ:]
                for i in range(x.max):
                    x[:] = i
                    req_vec.next = x
                    yield delay(10)

                    s = 0
                    v = 0
                    for k in range(NUM_REQ):
                        if x[k]==1:
                            s = k
                            v = 1
                            break
                    check(vec=(1<<s)*v, idx=s, vld=v)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut = getDut(arbiter_priority, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld)
            stm = stim()
            Simulation(dut, stm).run()
            del dut, stm




if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import random
import copy

def payload_generator(levels=0, dimensions=None, sequential=True, string=True, max_int=255, max_pkt_len=150, max_dim_size=3):
    """ Generates random data in a form of nested lists, to be used as a test packet payload
            levels - nesting levels:
                0 - list of integers (single packet payload)
                1 - list of lists of integers (list of packets)
                2 - list of lists of lists of integers (list of packets, each packet containing list of messages)
                etc.
            dimensions - specifies the size of the lists; the specification can be partial, as what is not specified will be chosen randomly
                if level=0:
                    dimensions=None - list of integers of random length in the range [1,max_pkt_len]
                    dimensions=5 - list of integers of length 5
                if level=1:
                    dimensions=None - list of random length in the range [1, max_dim_size], each element is an integer list of random length in the range [1, max_pkt_len]
                    dimensions=5 - list of length 5, each element is an integer list of random length in the range [1, max_pkt_len]
                    dimensions=[2, 3] - list of 2 elements, each element is an integer list, the first one of length 2, the second on of length 3
                if level=2:
                    dimensions=None - list of random length in the range [1, max_dim_size], each element is a list of random length in the range [1, max_dim_size], 
                                      each element of the second level lists is an integer list of random length in the range [1, max_pkt_len]
                    dimensions=5 - list of length 5, each element is a list of random length in the range [1, max_dim_size], each element of the second level lists 
                                   is an integer list of random length in the range [1, max_pkt_len]
                    dimensions=[2, 3] - list of 2 elements, each element is a list, the first one of length 2, the second on of length 3, each element of the second 
                                        level lists is an integer list of random length in the range [1, max_pkt_len]
                    dimensions=[[3, 5], [6, 7, 8]] - list of 2 elements, each element is a list, the first one of length 2, the second on of length 3, each element of the second 
                                        level lists is an integer list of length respectively 3, 5, 6, 7, 8
                if level=3:
                    etc.
                It is possible to specify dimensions partially, e.g dimensions=[[3, None], [6, 7, 8]] or dimensions=[None, [6, 7, 8]]. Dimensions that are not specified are 
                chosen randomly
            sequential - if True, the payload is a cyclic sequence if integers in the range [0, max_int]
                         if False, the payload is a sequence of random integers in the range [0, max_int]
            string - if True, the payload is a byte string, e.g. '\x04\x05\x06\x07' as each byte is in the range [0, min(255, max_int)]
                     if False, the payload is a list of integers in the range [0, max_int]
            max_int - Upper limit for the integer range for the payload
            max_pkt_len - Upper limit for the randomly chosen payload length
            max_dim_size - Upper limit for the randomly chosen dimension sizes
    """

    MAX_INT = max_int
    if string:
        MAX_INT = min(255, max_int)

    MAX_DIM_SIZE = max_dim_size
    MAX_PKT_LEN = max_pkt_len

    def next_i():
        ''' Generates next number from a cyclic integer sequence [0..MAX_INT]'''
        next_i.i = (next_i.i+1)%(MAX_INT+1)
        return next_i.i
    next_i.i = 0;

    def payload(length):
        ''' Generates payload of given length '''
        if (sequential) :
            pld = [next_i() for _ in xrange(length)]
        else:
            pld = [random.randint(0, MAX_INT) for _ in xrange(length)]
        if string:
            pld = str(bytearray(pld))
        return pld

    def next_level(level, pld):
        ''' Generates the next level of nested lists '''
        if level>0:
            # Next level of nested lists
            if pld==None:
                pld = random.randint(1, MAX_DIM_SIZE)
            if isinstance(pld, int):
                pld = pld*[None]
            if isinstance(pld, list):
                for i in range(len(pld)):
                    pld[i] = next_level(level-1, pld[i])
                return pld
            else:
                raise TypeError("Expected None, int or list, got {}: {}".format(type(pld), pld))
        elif level==0:
            # Generate payload
            if pld==None:
                pld = random.randint(1, MAX_PKT_LEN)
            if isinstance(pld, int):
                return payload(pld)
            else:
                raise TypeError("Expected None or int, got {}: {}".format(type(pld), pld))
        else:
            raise ValueError("Expected int>=0, got {}".format(level))

    pld = copy.deepcopy(dimensions)

    pld = next_level(levels, pld)

    return pld


if __name__ == '__main__':
    print payload_generator(levels=2, dimensions=[[3,4,5],[5,5,5]], sequential=False, string=False, max_int=8)
#     pass



import unittest

from myhdl import *
from myhdl_lib.stream import bytecount
import myhdl_lib.simulation as sim

import random
from math import log, floor, ceil


class TestBytecount(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testName(self):
        ''' BYTECOUNT: '''
        MAX_NUM_BYTES = 200
        BYTES_PER_WORD = [1,2,3,4,5]

        getDut = sim.DUTer()

        def testbench(BYTES_PER_WORD):
            MTY_WIDTH = int(floor(log(BYTES_PER_WORD,2))) + 1
            COUNT_WIDTH = int(floor(log(MAX_NUM_BYTES,2))) + 1

            clk = sim.Clock(val=0, period=10, units="ns")
            rst = sim.ResetSync(clk=clk, val=0, active=1)

            rx_vld = Signal(bool(0))
            rx_sop = Signal(bool(0))
            rx_eop = Signal(bool(0))
            rx_dat = Signal(intbv(0)[BYTES_PER_WORD*8:])
            rx_mty = Signal(intbv(0)[MTY_WIDTH:])
            count = Signal(intbv(0)[COUNT_WIDTH:])

            argl = {"rst":rst,
                    "clk":clk,
                    "rx_vld":rx_vld,
                    "rx_sop":rx_sop,
                    "rx_eop":rx_eop,
                    "rx_dat":rx_dat,
                    "rx_mty":rx_mty,
                    "count":count,
                    "MAX_BYTES":MAX_NUM_BYTES}

            dut = getDut(bytecount, **argl)
            clkgen = clk.gen()

            @instance
            def _stim():
                rx_vld.next = 0
                rx_sop.next = 0
                rx_eop.next = 0
                rx_dat.next = 0
                rx_mty.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                assert 0==count, "bytecount: expected {}, detected {}".format(0, count)

                for _ in range(10):
                    bytes = random.randint(1,MAX_NUM_BYTES)
                    words = int(ceil(float(bytes)/BYTES_PER_WORD))
                    empty_n = bytes%BYTES_PER_WORD
                    empty = BYTES_PER_WORD - empty_n
                    C = 0
                    for i in range(words):
                        rx_vld.next = 1
                        rx_sop.next = (i==0)
                        rx_eop.next = (i==words-1)
                        rx_dat.next = random.randint(0,2**BYTES_PER_WORD*8-1)
                        rx_mty.next = empty if i==words-1 else 0
                        C = C + (empty_n if i==words-1 else BYTES_PER_WORD)
                        yield clk.posedge
                        yield delay(1)
                        assert C==count, "bytecount: expected {}, detected {}".format(C, count)
                        # Gap
                        rx_vld.next = 0
                        for _ in range(random.randint(0,3)):
                            yield clk.posedge

                yield clk.posedge
                raise StopSimulation
            return instances()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for BPW in BYTES_PER_WORD:
                tb = testbench(BPW)
                Simulation(tb).run()
                del tb


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import unittest

from myhdl import *
from myhdl_lib.handshake import hs_arbmux
import myhdl_lib.simulation as sim


class TestArbMux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    @staticmethod
    def hs_arbmux_top(rst, clk, i0_rdy, i0_vld, i1_rdy, i1_vld, i2_rdy, i2_vld, o_rdy, o_vld, sel, ARBITER_TYPE):
        ''' Needed when hs_arbmux is co-simulated as top level'''

        hsi0_rdy = Signal(bool(0))
        hsi0_vld = Signal(bool(0))
        hsi1_rdy = Signal(bool(0))
        hsi1_vld = Signal(bool(0))
        hsi2_rdy = Signal(bool(0))
        hsi2_vld = Signal(bool(0))
        hso_rdy = Signal(bool(0))
        hso_vld = Signal(bool(0))

        @always_comb
        def _assign():
            i0_rdy.next = hsi0_rdy
            i1_rdy.next = hsi1_rdy
            i2_rdy.next = hsi2_rdy

            hsi0_vld.next = i0_vld
            hsi1_vld.next = i1_vld
            hsi2_vld.next = i2_vld

            hso_rdy.next = o_rdy
            o_vld.next = hso_vld

        ls_hsi = [(hsi0_rdy, hsi0_vld), (hsi1_rdy, hsi1_vld), (hsi2_rdy, hsi2_vld)]
        hso = (hso_rdy, hso_vld)

        _inst = hs_arbmux(rst=rst, clk=clk, ls_hsi=ls_hsi, hso=hso, sel=sel, ARBITER_TYPE=ARBITER_TYPE)

        return instances()

    def testArbMux3Prio(self):
        "HS_ARBMUX: 3 inputs, priority arbiter"

        NUM_INPUTS = 3

        ls_hsi_rdy = [Signal(bool(0)) for _ in range(NUM_INPUTS)]
        ls_hsi_vld = [Signal(bool(0)) for _ in range(NUM_INPUTS)]
        ls_hsi = zip(ls_hsi_rdy, ls_hsi_vld)

        hso_rdy = Signal(bool(0))
        hso_vld = Signal(bool(0))
        hso = (hso_rdy, hso_vld)

        sel = Signal(intbv(0)[NUM_INPUTS:])

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        argl = {"rst":None, "clk":None,
                "i0_rdy":ls_hsi[0][0], "i0_vld":ls_hsi[0][1],
                "i1_rdy":ls_hsi[1][0], "i1_vld":ls_hsi[1][1],
                "i2_rdy":ls_hsi[2][0], "i2_vld":ls_hsi[2][1],
                "o_rdy":hso[0], "o_vld":hso[1], "sel":sel, "ARBITER_TYPE":"priority"}

        def stim():
            @instance
            def _inst():
                for i in range(NUM_INPUTS):
                    ls_hsi_vld[i].next = 0
                hso_rdy.next = 0
                yield delay(10)
                for i in range(NUM_INPUTS):
                    assert 0==ls_hsi_rdy[i], "hsi_rdy[{}]: expected {}, detected {}".format(i, 0, ls_hsi_rdy[i])
                assert 0==hso_vld, "hso_vld: expected {}, detected {}".format( 0, hso_vld)
                assert 0==sel, "sel: expected {}, detected {}".format( 0, sel)

                x = intbv(0)[NUM_INPUTS+1:]
                for i in range(x.max):
                    x[:] = i
                    for k in range(NUM_INPUTS):
                        ls_hsi_vld[k].next = x[k]
                    hso_rdy.next = x[NUM_INPUTS]

                    yield delay(10)

                    s = 0
                    for i in range(NUM_INPUTS):
                        if x[i]==1:
                            s = i
                            break

                    assert s==sel, "sel: expecte {}, dectected {}".format(s, sel)

                    for k in range(NUM_INPUTS):
                        if (k==s) and (x[NUM_INPUTS]==1):
                            assert 1==ls_hsi_rdy[k], "hsi_rdy[{}]: expected {}, detected {}".format(k, 1, ls_hsi_rdy[k])
                        else:
                            assert 0==ls_hsi_rdy[k], "hsi_rdy[{}]: expected {}, detected {}".format(k, 0, ls_hsi_rdy[k])

                    if x[s]==1:
                        assert 1==hso_vld, "hso_vld: expected {}, detected {}".format(1, hso_vld)
                    else:
                        assert 0==hso_vld, "hso_vld: expected {}, detected {}".format(0, hso_vld)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            clkgen = clk.gen()

            dut=getDut(self.hs_arbmux_top, **argl)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


    def testArbMux3RR(self):
        "HS_ARBMUX: 3 inputs, roundrobin arbiter"

        NUM_INPUTS = 3

        ls_hsi_rdy = [Signal(bool(0)) for _ in range(NUM_INPUTS)]
        ls_hsi_vld = [Signal(bool(0)) for _ in range(NUM_INPUTS)]
        ls_hsi = zip(ls_hsi_rdy, ls_hsi_vld)

        hso_rdy = Signal(bool(0))
        hso_vld = Signal(bool(0))
        hso = (hso_rdy, hso_vld)

        sel = Signal(intbv(0)[NUM_INPUTS:])

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        argl = {"rst":rst, "clk":clk,
                "i0_rdy":ls_hsi[0][0], "i0_vld":ls_hsi[0][1],
                "i1_rdy":ls_hsi[1][0], "i1_vld":ls_hsi[1][1],
                "i2_rdy":ls_hsi[2][0], "i2_vld":ls_hsi[2][1],
                "o_rdy":hso[0], "o_vld":hso[1], "sel":sel, "ARBITER_TYPE":"roundrobin"}

        def stim():
            @instance
            def _inst():
                for i in range(NUM_INPUTS):
                    ls_hsi_vld[i].next = 0
                hso_rdy.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                for i in range(NUM_INPUTS):
                    assert 0==ls_hsi_rdy[i], "hsi_rdy[{}]: expected {}, detected {}".format(i, 0, ls_hsi_rdy[i])
                assert 0==hso_vld, "hso_vld: expected {}, detected {}".format( 0, hso_vld)
                assert 0==sel, "sel: expected {}, detected {}".format( 0, sel)

                prio = NUM_INPUTS-1
                x = intbv(0)[NUM_INPUTS+1:]
                for i in range(x.max):
                    x[:] = i
                    for k in range(NUM_INPUTS):
                        ls_hsi_vld[k].next = x[k]
                    hso_rdy.next = x[NUM_INPUTS]

                    yield clk.posedge

                    s = 0
                    for p in range(NUM_INPUTS):
                        j = (prio+p+1)%NUM_INPUTS
                        if x[j]==1:
                            s = j
                            if hso_rdy and hso_vld:
                                prio = j
                            break

                    assert s==sel, "sel: expecte {}, dectected {}".format(s, sel)

                    for k in range(NUM_INPUTS):
                        if (k==s) and (x[NUM_INPUTS]==1):
                            assert 1==ls_hsi_rdy[k], "hsi_rdy[{}]: expected {}, detected {}".format(k, 1, ls_hsi_rdy[k])
                        else:
                            assert 0==ls_hsi_rdy[k], "hsi_rdy[{}]: expected {}, detected {}".format(k, 0, ls_hsi_rdy[k])

                    if x[s]==1:
                        assert 1==hso_vld, "hso_vld: expected {}, detected {}".format(1, hso_vld)
                    else:
                        assert 0==hso_vld, "hso_vld: expected {}, detected {}".format(0, hso_vld)

                yield clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            clkgen = clk.gen()

            dut=getDut(self.hs_arbmux_top, **argl)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import unittest

from myhdl import *
from myhdl_lib.mux import demux
import myhdl_lib.simulation as sim

import random


class TestDemux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]


    def testDemux5(self):
        ''' DEMUX: 5 outputs '''
        DMAX = 100
        NUM_OUTPUTS = 5

        def demux_top(sel, di, do_0, do_1, do_2, do_3, do_4):
            ''' Needed when demux is co-simulated as top level'''
            ls_do = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_OUTPUTS)]
            @always_comb
            def _assign():
                do_0.next = ls_do[0]
                do_1.next = ls_do[1]
                do_2.next = ls_do[2]
                do_3.next = ls_do[3]
                do_4.next = ls_do[4]

            inst = demux(sel=sel, di=di, ls_do=ls_do)
            return instances()

        sel = Signal(intbv(0,min=0,max=NUM_OUTPUTS))
        di = Signal(intbv(0,min=0,max=DMAX))
        ls_do = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_OUTPUTS)]

        argl = {"sel":sel, "di":di, "do_0":ls_do[0], "do_1":ls_do[1], "do_2":ls_do[2], "do_3":ls_do[3], "do_4":ls_do[4]}


        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(NUM_OUTPUTS):
                    di.next = 0
                    sel.next = 0
                yield delay(10)

                for _ in range(10):
                    d = [random.randint(0,DMAX-1) for _ in range(NUM_OUTPUTS)]

                    for s in range(NUM_OUTPUTS):
                        di.next = d[s]
                        sel.next = s
                        yield delay(10)

                        for i in range(NUM_OUTPUTS):
                            if s == i:
                                assert d[s] == ls_do[i], "Mux output {} (sel={}): expected {}, detected {}".format(i, sel, d[s], ls_do[i])
                            else:
                                assert 0 == ls_do[i], "Mux output {} (sel={}): expected {}, detected {}".format(i, sel, d[s], ls_do[i])

                yield delay(10)
                raise StopSimulation
            return _inst


        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(demux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm


    def testDemux2(self):
        ''' DEMUX: 2 outputs, boolean Select '''
        DMAX = 100
        NUM_OUTPUTS = 2

        def demux_top(sel, di, do_0, do_1):
            ''' Needed when demux is co-simulated as top level'''
            ls_do = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_OUTPUTS)]
            @always_comb
            def _assign():
                do_0.next = ls_do[0]
                do_1.next = ls_do[1]

            inst = demux(sel=sel, di=di, ls_do=ls_do)
            return instances()

        sel = Signal(bool(0))
        di = Signal(intbv(0,min=0,max=DMAX))
        ls_do = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_OUTPUTS)]

        argl = {"sel":sel, "di":di, "do_0":ls_do[0], "do_1":ls_do[1]}


        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(NUM_OUTPUTS):
                    di.next = 0
                    sel.next = 0
                yield delay(10)

                for _ in range(10):
                    d = [random.randint(0,DMAX-1) for _ in range(NUM_OUTPUTS)]

                    for s in range(NUM_OUTPUTS):
                        di.next = d[s]
                        sel.next = s
                        yield delay(10)

                        for i in range(NUM_OUTPUTS):
                            if s == i:
                                assert d[s] == ls_do[i], "Mux output {} (sel={}): expected {}, detected {}".format(i, sel, d[s], ls_do[i])
                            else:
                                assert 0 == ls_do[i], "Mux output {} (sel={}): expected {}, detected {}".format(i, sel, 0, ls_do[i])

                yield delay(10)
                raise StopSimulation
            return _inst


        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(demux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Mar 1, 2014

@author: nkavaldj
'''
class sfifo_beh:
    ''' Speculative FIFO: behavioral model '''

    COMMIT = 0
    DISCARD = 1
    DISCARD_COMMIT = 2

    def __init__(self, depth, afull_th=None, aempty_th=None):
        self.depth = depth
        self.data = [] # Data (written, committed, not read )
        self.data_sw = [] # Speculatively written data
        self.data_sr = [] # Speculatively read data
        self.afull_th = afull_th if (afull_th != None) else depth//2
        self.aempty_th = aempty_th if (aempty_th != None) else depth//2
        self.ovf = False
        self.udf = False
        self.count_max = 0

    def reset(self):
        self.__init__(self.depth, self.afull_th, self.aempty_th)

    def isFull(self):
        return self.getCount() == self.depth

    def isEmpty(self):
        return len(self.data) == 0

    def isAFull(self):
        return (self.depth - self.getCount()) <= self.afull_th

    def isAEmpty(self):
        return len(self.data) <= self.aempty_th

    def isOvf(self):
        return self.ovf

    def isUdf(self):
        return self.udf

    def getCount(self):
        return len(self.data_sr) + len(self.data) + len(self.data_sw)

    def getCountMax(self):
        return self.count_max

    def getDout(self):
        return self.data[0]


    def _rcmd(self, cmd):
        if (cmd == self.DISCARD or cmd == self.DISCARD_COMMIT):
            self.data = self.data_sr + self.data
            self.data_sr = []
        elif (cmd == self.COMMIT):
            self.data_sr = []

    def _wcmd(self, cmd):
        if (cmd == self.DISCARD or cmd == self.DISCARD_COMMIT):
            self.data_sw = []
        elif (cmd == self.COMMIT):
            self.data.extend(self.data_sw)
            self.data_sw = []

    def command(self, wcmd=None, rcmd=None):
        self._wcmd(wcmd)
        self._rcmd(rcmd)

    def write(self, val, wcmd=None, rcmd=None):
        if (not self.isFull()):
            self.data_sw.append(val)
        else:
            self.ovf = True

        self._wcmd(wcmd)
        self._rcmd(rcmd)

        if (self.getCount() > self.count_max):
            self.count_max = self.getCount()

    def read(self, wcmd=None, rcmd=None):
        x = None
        if (not self.isEmpty()):
            x = self.data.pop(0)
            self.data_sr.append(x)
        else:
            self.udf = True

        self._wcmd(wcmd)
        self._rcmd(rcmd)

        return x

    def write_read(self, val, wcmd=None, rcmd=None):
        full = self.isFull()
        empty = self.isEmpty()

        if (not full):
            self.data_sw.append(val)
        else:
            self.ovf = True

        self._wcmd(wcmd)

        x = None
        if (not empty):
            x = self.data.pop(0)
            self.data_sr.append(x)
        else:
            self.udf = True

        self._rcmd(rcmd)

        if (self.getCount() > self.count_max):
            self.count_max = self.getCount()
        return x


    def status(self):
        return "FULL = {:}, EMPTY = {:}, COUNT = {:3}, AFULL = {:}, AEMPTY = {:}, OVF = {:}, UDF = {:}, COUNT_MAX = {:}, DOUT = {:}".format(int(self.isFull()), int(self.isEmpty()), int(self.getCount()), int(self.isAFull()), int(self.isAEmpty()), int(self.isOvf()), int(self.isUdf()), self.getCountMax(), self.data[0] if len(self.data)>0 else None)



        return "FULL = {:}, EMPTY = {:}, COUNT = {:3}, AFULL = {:}, AEMPTY = {:}, OVF = {:}, UDF = {:}, COUNT_MAX = {:}, DOUT = {:}".format(int(self.isFull()), int(self.isEmpty()), int(self.count()), int(self.isAFull()), int(self.isAEmpty()), int(self.isOvf()), int(self.isUdf()), self.maxCount(), self.data[0] if len(self.data)>0 else None)


from myhdl import *
from mem   import ram_sdp_ar


def fifo_async(wrst, rrst, wclk, rclk, wfull, we, wdata, rempty, re, rdata, depth=None, width=None):
    ''' Asynchronous FIFO

        Implements the design described in:
            Clifford E. Cummings, "Simulation and Synthesis Techniques for Asynchronous FIFO Design," SNUG 2002 (Synopsys
            Users Group Conference, San Jose, CA, 2002) User Papers, March 2002, Section TB2, 2nd paper. Also available at
            www.sunburst-design.com/papers

        Write side interface:
            wrst   - reset 
            wclk   - clock
            wfull  - full flag, immediate set on 'write', delayed clear on 'read' due to clock domain synchronization
            we     - write enable
            wdata  - write data
        Read side interface
            rrst   - reset
            cclk   - clock
            rempty - empty flag, immediate set on 'read', delayed clear on 'write' due to clock domain synchronization
            re     - read enable
            rdata  - read data
        Parameters
            depth - fifo depth. If not set, default 4 is used. Must be >=4. Must be power of 2
            width - data width. If not set, data with equals len(wdata). Can be [0,1,2,3...)
                    It is possible to instantiate a fifo with data width 0 (no data) if width=0 or width=None and wdata=None
    '''

    if (width == None):
        width = 0
        if wdata is not None:
            width = len(wdata)

    if (depth == None):
        depth = 4
    assert depth >= 4, "Fifo_async parameter 'depth' must be >= 4 , detected depth={}".format(depth)
    assert (depth & (depth-1)) == 0, "Fifo_async parameter 'depth' must be 2**n, detected depth={}".format(depth)


    full_flg        = Signal(bool(1))
    empty_flg       = Signal(bool(1))
    full_val        = Signal(bool(1))
    empty_val       = Signal(bool(1))
    we_safe         = Signal(bool(0))
    re_safe         = Signal(bool(0))
    rd_ptr          = Signal(intbv(0, min=0, max=depth))
    wr_ptr          = Signal(intbv(0, min=0, max=depth))

    WIDTH = len(rd_ptr)

    rd_ptr_bin          = Signal(intbv(0)[WIDTH+1:])
    rd_ptr_bin_new      = Signal(intbv(0)[WIDTH+1:])
    rd_ptr_gray         = Signal(intbv(0)[WIDTH+1:])
    rd_ptr_gray_new     = Signal(intbv(0)[WIDTH+1:])
    rd_ptr_gray_sync1   = Signal(intbv(0)[WIDTH+1:])
    rd_ptr_gray_sync2   = Signal(intbv(0)[WIDTH+1:])
    wr_ptr_bin          = Signal(intbv(0)[WIDTH+1:])
    wr_ptr_bin_new      = Signal(intbv(0)[WIDTH+1:])
    wr_ptr_gray         = Signal(intbv(0)[WIDTH+1:])
    wr_ptr_gray_new     = Signal(intbv(0)[WIDTH+1:])
    wr_ptr_gray_sync1   = Signal(intbv(0)[WIDTH+1:])
    wr_ptr_gray_sync2   = Signal(intbv(0)[WIDTH+1:])

    @always_comb
    def safe_read_write():
        wfull.next      = full_flg
        rempty.next     = empty_flg
        we_safe.next    = we and not full_flg
        re_safe.next    = re and not empty_flg

    @always(wclk.posedge)
    def sync_r2w():
        ''' Read-domain to write-domain synchronizer '''
        if (wrst):
            rd_ptr_gray_sync1.next = 0
            rd_ptr_gray_sync2.next = 0
        else:
            rd_ptr_gray_sync1.next = rd_ptr_gray
            rd_ptr_gray_sync2.next = rd_ptr_gray_sync1

    @always(rclk.posedge)
    def sync_w2r():
        ''' Write-domain to read-domain synchronizer '''
        if (rrst):
            wr_ptr_gray_sync1.next = 0
            wr_ptr_gray_sync2.next = 0
        else:
            wr_ptr_gray_sync1.next = wr_ptr_gray
            wr_ptr_gray_sync2.next = wr_ptr_gray_sync1

    @always_comb
    def bin_comb():
        wr_ptr_bin_new.next = wr_ptr_bin
        rd_ptr_bin_new.next = rd_ptr_bin
        if (we_safe):
            wr_ptr_bin_new.next = (wr_ptr_bin + 1) % wr_ptr_bin.max
        if (re_safe):
            rd_ptr_bin_new.next = (rd_ptr_bin + 1) % rd_ptr_bin.max

    @always_comb
    def gray_comb():
        wr_ptr_gray_new.next    = (wr_ptr_bin_new >> 1) ^ wr_ptr_bin_new
        rd_ptr_gray_new.next    = (rd_ptr_bin_new >> 1) ^ rd_ptr_bin_new

    @always_comb
    def full_empty_comb():
        empty_val.next  = (rd_ptr_gray_new == wr_ptr_gray_sync2)
        full_val.next   = (wr_ptr_gray_new[WIDTH] != rd_ptr_gray_sync2[WIDTH]) and \
                          (wr_ptr_gray_new[WIDTH-1] != rd_ptr_gray_sync2[WIDTH-1]) and \
                          (wr_ptr_gray_new[WIDTH-1:] == rd_ptr_gray_sync2[WIDTH-1:])


    @always(wclk.posedge)
    def wptr_proc():
        if (wrst):
            wr_ptr_bin.next     = 0
            wr_ptr_gray.next    = 0
            full_flg.next       = 0
        else:
            wr_ptr_bin.next     = wr_ptr_bin_new
            wr_ptr_gray.next    = wr_ptr_gray_new
            full_flg.next       = full_val

    @always(rclk.posedge)
    def rptr_proc():
        if (rrst):
            rd_ptr_bin.next     = 0
            rd_ptr_gray.next    = 0
            empty_flg.next      = 1
        else:
            rd_ptr_bin.next     = rd_ptr_bin_new
            rd_ptr_gray.next    = rd_ptr_gray_new
            empty_flg.next      = empty_val


    if width>0:
        #===========================================================================
        # Memory instance
        #===========================================================================
        mem_we      = Signal(bool(0))
        mem_addrw   = Signal(intbv(0, min=0, max=depth))
        mem_addrr   = Signal(intbv(0, min=0, max=depth))
        mem_di      = Signal(intbv(0)[width:0])
        mem_do      = Signal(intbv(0)[width:0])


        # RAM: Simple-Dual-Port, Asynchronous read
        mem = ram_sdp_ar(   clk     = wclk,
                            we      = mem_we,
                            addrw   = mem_addrw,
                            addrr   = mem_addrr,
                            di      = mem_di,
                            do      = mem_do )

        @always_comb
        def mem_connect():
            mem_we.next         = we_safe
            mem_addrw.next      = wr_ptr_bin[WIDTH:]
            mem_addrr.next      = rd_ptr_bin[WIDTH:]
            mem_di.next         = wdata
            rdata.next          = mem_do


    return instances()


def convert(depth=256, width=8):
    wrst, rrst, wclk, rclk   = [Signal(bool(0)) for _ in range(4)]
    wdata, rdata  = [Signal(intbv(0)[width:]) for _ in range(2)]
    re, we, rempty, wfull = [Signal(bool(0)) for _ in range(4)]

    toVerilog(fifo_async, wrst, rrst, wclk, rclk, wfull, we, wdata, rempty, re, rdata, depth=depth, width=width)



if __name__ == '__main__':
    convert()


import unittest

from myhdl import *
from myhdl_lib.mux import ls_demux
import myhdl_lib.simulation as sim

import random


class TestLsDemux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testLsDemux5(self):
        ''' LS_DEMUX: 5 outputs '''
        DMAX = 100
        NUM_OUTPUTS = 5

        def ls_demux_top(sel, 
                       di_0, di_1, di_2,
                       do_00, do_01, do_02,
                       do_10, do_11, do_12,
                       do_20, do_21, do_22,
                       do_30, do_31, do_32,
                       do_40, do_41, do_42):
            ''' Needed when ls_demux is co-simulated as top level'''
            ls_di_0 = Signal(bool(0))
            ls_di_1 = Signal(intbv(0,min=0,max=DMAX))
            ls_di_2 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_do_00 = Signal(bool(0))
            lsls_do_01 = Signal(intbv(0,min=0,max=DMAX))
            lsls_do_02 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_do_10 = Signal(bool(0))
            lsls_do_11 = Signal(intbv(0,min=0,max=DMAX))
            lsls_do_12 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_do_20 = Signal(bool(0))
            lsls_do_21 = Signal(intbv(0,min=0,max=DMAX))
            lsls_do_22 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_do_30 = Signal(bool(0))
            lsls_do_31 = Signal(intbv(0,min=0,max=DMAX))
            lsls_do_32 = Signal(intbv(0,min=-DMAX,max=DMAX))

            lsls_do_40 = Signal(bool(0))
            lsls_do_41 = Signal(intbv(0,min=0,max=DMAX))
            lsls_do_42 = Signal(intbv(0,min=-DMAX,max=DMAX))

            @always_comb
            def _assign():
                ls_di_0.next = di_0
                ls_di_1.next = di_1
                ls_di_2.next = di_2

                do_00.next = lsls_do_00
                do_01.next = lsls_do_01
                do_02.next = lsls_do_02
                do_10.next = lsls_do_10
                do_11.next = lsls_do_11
                do_12.next = lsls_do_12
                do_20.next = lsls_do_20
                do_21.next = lsls_do_21
                do_22.next = lsls_do_22
                do_30.next = lsls_do_30
                do_31.next = lsls_do_31
                do_32.next = lsls_do_32
                do_40.next = lsls_do_40
                do_41.next = lsls_do_41
                do_42.next = lsls_do_42

            lsls_do = [[lsls_do_00, lsls_do_01, lsls_do_02],
                       [lsls_do_10, lsls_do_11, lsls_do_12],
                       [lsls_do_20, lsls_do_21, lsls_do_22],
                       [lsls_do_30, lsls_do_31, lsls_do_32],
                       [lsls_do_40, lsls_do_41, lsls_do_42]]

            ls_di = [ls_di_0, ls_di_1, ls_di_2]

            _ls_demux = ls_demux(sel=sel, ls_di=ls_di, lsls_do=lsls_do)
            return instances()


        sel = Signal(intbv(0,min=0,max=NUM_OUTPUTS))

        ls_di = [Signal(bool(0)),
                 Signal(intbv(0,min=0,max=DMAX)),
                 Signal(intbv(0,min=-DMAX,max=DMAX))]

        lsls_do = [[Signal(bool(0)),
                    Signal(intbv(0,min=0,max=DMAX)),
                    Signal(intbv(0,min=-DMAX,max=DMAX))] for _ in range(NUM_OUTPUTS)]

        argl = {"sel":sel, 
                "di_0":ls_di[0],"di_1":ls_di[1],"di_2":ls_di[2],
                "do_00":lsls_do[0][0], "do_01":lsls_do[0][1], "do_02":lsls_do[0][2],
                "do_10":lsls_do[1][0], "do_11":lsls_do[1][1], "do_12":lsls_do[1][2],
                "do_20":lsls_do[2][0], "do_21":lsls_do[2][1], "do_22":lsls_do[2][2],
                "do_30":lsls_do[3][0], "do_31":lsls_do[3][1], "do_32":lsls_do[3][2],
                "do_40":lsls_do[4][0], "do_41":lsls_do[4][1], "do_42":lsls_do[4][2]}

        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(3):
                    ls_di[i].next = 0
                yield delay(10)

                for _ in range(2):
                    d = [[random.randint(0,1),
                          random.randint(0,DMAX-1),
                          random.randint(-DMAX,DMAX-1)] for _ in range(NUM_OUTPUTS)]

                    for s in range(NUM_OUTPUTS):
                        for j in range(3):
                            ls_di[j].next = d[s][j]
                        sel.next = s
                        yield delay(10)

                        for o in range(NUM_OUTPUTS):
                            if s == o:
                                assert d[s][0] == lsls_do[o][0], "Mux output {} (sel={}): expected {}, detected {}".format(o, sel, d[s], [int(x) for x in lsls_do[o]])
                                assert d[s][1] == lsls_do[o][1], "Mux output {} (sel={}): expected {}, detected {}".format(o, sel, d[s], [int(x) for x in lsls_do[o]])
                                assert d[s][2] == lsls_do[o][2], "Mux output {} (sel={}): expected {}, detected {}".format(o, sel, d[s], [int(x) for x in lsls_do[o]])
                            else:
                                assert 0 == lsls_do[o][0], "Mux output {} (sel={}): expected {}, detected {}".format(o, sel, 3*[0], [int(x) for x in lsls_do[o]])
                                assert 0 == lsls_do[o][1], "Mux output {} (sel={}): expected {}, detected {}".format(o, sel, 3*[0], [int(x) for x in lsls_do[o]])
                                assert 0 == lsls_do[o][2], "Mux output {} (sel={}): expected {}, detected {}".format(o, sel, 3*[0], [int(x) for x in lsls_do[o]])

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(ls_demux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


from myhdl import *


def mux(sel, ls_di, do):
    """ Multiplexes a list of input signals to an output signal
            do = sl_di[sel]

            sel - select index
            ls_di - list of input signals
            do - output signals
            
    """
    N = len(ls_di)
    @always_comb
    def _mux():
        do.next = 0
        for i in range(N):
            if i == sel:
                do.next = ls_di[i]
    return _mux


def demux(sel, di, ls_do):
    """ Demultiplexes an input signal to a list of output signals
            ls_do[sel] =  di

            sel - select index
            di - input signal
            ls_do - list of output signals
    """
    N = len(ls_do)
    @always_comb
    def _demux():
        for i in range(N):
            ls_do[i].next = 0
            if i == sel:
                ls_do[i].next = di
    return _demux


def ls_mux(sel, lsls_di, ls_do):
    """ Multiplexes a list of input signal structures to an output structure. 
        A structure is represented by a list of signals: [signal_1, signal_2, ..., signal_n]
            ls_do[0] = lsls_di[sel][0]
            ls_do[1] = lsls_di[sel][1]
            ...
            ls_do[n] = lsls_di[sel][n]

            sel - select index
            lsls_di - list of input signal structures: [[sig, sig, ..., sig], [sig, sig, ..., sig], ..., [sig, sig, ..., sig]]
            ls_do - output signal structure: [sig, sig, ..., sig]
    """
    N = len(ls_do)
    lsls_in = [list(x) for x in zip(*lsls_di)]
    return [mux(sel, lsls_in[i], ls_do[i]) for i in range(N)]


def ls_demux(sel, ls_di, lsls_do):
    """ Demultiplexes an input signal structure to list of output structures. 
        A structure is represented by a list of signals: [signal_1, signal_2, ..., signal_n]
            lsls_do[sel][0] = ls_di[0]
            lsls_do[sel][1] = ls_di[1]
            ...
            lsls_do[sel][n] = ls_di[n]

            sel - select index
            ls_di - input signal structure: [sig, sig, ..., sig]
            lsls_do - list of output signal structures: [[sig, sig, ..., sig], [sig, sig, ..., sig], ..., [sig, sig, ..., sig]]
    """
    N = len (ls_di)
    lsls_out = [list(x) for x in zip(*lsls_do)]
    return [demux(sel, ls_di[i], lsls_out[i])for i in range(N)]


def bitslice_select(offset, bv_di, bv_do):
    ''' Selects a bit-slice from a bit-vector
            offset - (i) bit offset of the slice
            bv_di  - (i) bit vector where the slice is taken from
            bv_do  - (o) selected slice; the length of this bit-vector defines the number of bit in the slice

            bv_do = bv_di[len(bv_do)+offset:offset]
    '''
    LEN_I = len(bv_di)
    LEN_O = len(bv_do)

    assert LEN_I >= LEN_O, "bitslice_select: expects len(bv_di) >= len(bv_do), but len(bv_di)={}, len(bv_do)".format(LEN_I, LEN_O)

    OFFSET_MAX = LEN_I - LEN_O + 1

    @always_comb
    def _slice():
        bv_do.next = 0
        for i in range(OFFSET_MAX):
            if i==offset:
                for b in range(LEN_O):
                    bv_do.next[b] = bv_di[i+b]

    return _slice


def byteslice_select(offset, bv_di, bv_do):
    ''' Selects a slice of length 8*n aligned on a byte from a bit-vector
            offset - (i) byte offset of the slice
            bv_di  - (i) bit vector where the slice is taken from; must len(bv_di) = 8*m
            bv_do  - (o) selected slice; must len(bv_do) = 8*n, n<=m; len(bv_do) defines the number of bit in the slice
    '''
    LEN_I = len(bv_di)
    LEN_O = len(bv_do)

    assert (LEN_I % 8)==0, "byteslice_select: expects len(bv_di)=8*x, but len(bv_di)={} bits".format(LEN_I)
    assert (LEN_O % 8)==0, "byteslice_select: expects len(bv_do)=8*x, but len(bv_do)={} bits".format(LEN_O)

    bit_offset = Signal(intbv(0)[len(offset)+3:])

    @always_comb
    def _offset():
        bit_offset.next = offset << 3

    _slice = bitslice_select(bit_offset, bv_di, bv_do)

    return _offset, _slice

if __name__ == '__main__':
    pass


import unittest

from myhdl import *
from myhdl_lib.handshake import hs_join
import myhdl_lib.simulation as sim

class TestHSJoin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testHSJoin3(self):
        ''' HS_JOIN: 3 inputs '''
        NUM_INPUTS = 3

        def hs_join_top(i0_rdy, i0_vld, i1_rdy, i1_vld, i2_rdy, i2_vld, o_rdy, o_vld):
            ''' Needed when hs_join is co-simulated as top level'''

            hsi0_rdy = Signal(bool(0))
            hsi0_vld = Signal(bool(0))
            hsi1_rdy = Signal(bool(0))
            hsi1_vld = Signal(bool(0))
            hsi2_rdy = Signal(bool(0))
            hsi2_vld = Signal(bool(0))
            hso_rdy = Signal(bool(0))
            hso_vld = Signal(bool(0))

            @always_comb
            def _assign():
                i0_rdy.next = hsi0_rdy
                i1_rdy.next = hsi1_rdy
                i2_rdy.next = hsi2_rdy

                hsi0_vld.next = i0_vld
                hsi1_vld.next = i1_vld
                hsi2_vld.next = i2_vld

                hso_rdy.next = o_rdy
                o_vld.next = hso_vld

            ls_hsi = [(hsi0_rdy, hsi0_vld), (hsi1_rdy, hsi1_vld), (hsi2_rdy, hsi2_vld)]
            hso = (hso_rdy, hso_vld)

            _inst = hs_join(ls_hsi=ls_hsi, hso=hso)

            return instances()

        ls_hsi_rdy = [Signal(bool(0)) for _ in range(NUM_INPUTS)]
        ls_hsi_vld = [Signal(bool(0)) for _ in range(NUM_INPUTS)]
        ls_hsi = zip(ls_hsi_rdy, ls_hsi_vld)

        hso_rdy = Signal(bool(0))
        hso_vld = Signal(bool(0))
        hso = (hso_rdy, hso_vld)

        argl = {"i0_rdy":ls_hsi[0][0], "i0_vld":ls_hsi[0][1],
                "i1_rdy":ls_hsi[1][0], "i1_vld":ls_hsi[1][1],
                "i2_rdy":ls_hsi[2][0], "i2_vld":ls_hsi[2][1],
                "o_rdy":hso[0], "o_vld":hso[1]}

        def stim():
            @instance
            def _inst():
                for i in range(NUM_INPUTS):
                    ls_hsi_vld[i].next = 0
                hso_rdy.next = 0
                yield delay(10)
                for i in range(NUM_INPUTS):
                    assert 0==ls_hsi_rdy[i], "hsi_rdy[{}]: expected {}, detected {}".format(i, 0, ls_hsi_rdy[i])
                assert 0==hso_vld, "hso_vld: expected {}, detected {}".format( 0, hso_vld)

                x = intbv(0)[NUM_INPUTS+1:]
                for i in range(x.max):
                    x[:] = i
                    for k in range(NUM_INPUTS):
                        ls_hsi_vld[k].next = x[k]
                    hso_rdy.next = x[NUM_INPUTS]

                    yield delay(10)

                    if i==x.max-1:
                        for k in range(NUM_INPUTS):
                            assert 1==ls_hsi_rdy[k], "hsi_rdy[{}]: expected {}, detected {}".format(k, 1, ls_hsi_rdy[k])
                    else:
                        for k in range(NUM_INPUTS):
                            assert 0==ls_hsi_rdy[k], "hsi_rdy[{}]: expected {}, detected {}".format(k, 0, ls_hsi_rdy[k])

                    if (i==x.max-1) or (i==x.max/2-1):
                        assert 1==hso_vld, "hso_vld: expected {}, detected {}".format(1, hso_vld)
                    else:
                        assert 0==hso_vld, "hso_vld: expected {}, detected {}".format(0, hso_vld)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(hs_join_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import unittest

from myhdl import *
from myhdl_lib.pipeline_control import pipeline_control
import myhdl_lib.simulation as sim


class TestPipelineControl(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def setUp(self):
        self.clk = sim.Clock(val=0, period=10, units="ns")
        self.rst = sim.ResetSync(clk=self.clk, val=0, active=1)
        self.clkgen = self.clk.gen()

        self.rx_rdy = Signal(bool(0))
        self.rx_vld = Signal(bool(0))
        self.tx_rdy = Signal(bool(0))
        self.tx_vld = Signal(bool(0))


    def tearDown(self):
        pass

    def reset(self):
        yield self.rst.pulse(10)
        self.rst.next = 1
        for _ in range (5): 
            yield self.clk.posedge 
        self.rst.next = 0

    def pipe_state_check(self, rx_rdy, tx_vld, stage_en):
        yield delay(1)
        assert rx_rdy==self.rx_rdy, "RX_RDY: expected {}, detected {}".format(rx_rdy, self.rx_rdy)
        assert tx_vld==self.tx_vld, "TX_VLD: expected {}, detected {}".format(tx_vld, self.tx_vld)
        assert stage_en==self.en, "STAGE_EN: expected {}, detected {}".format(bin(stage_en), bin(self.en))

    def pipe_drive(self, rx_vld, tx_rdy, stop_rx, stop_tx):
        self.rx_vld.next = rx_vld
        self.tx_rdy.next = tx_rdy
        self.stop_rx.next = stop_rx
        self.stop_tx.next = stop_tx
        yield self.clk.posedge
        self.rx_vld.next = 0
        self.tx_rdy.next = 0
        self.stop_rx.next = 0
        self.stop_tx.next = 0

    def testSingleWrite(self):
        ''' PIPE_CTRL: Single data '''
        DEPTH = [1,2,3,7,8,9]

        def stim(DEPTH):
            @instance
            def _inst():
                yield self.reset()
                yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                yield self.clk.posedge

                for _ in range(3):
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    self.rx_vld.next = 1
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=1)
                    yield self.clk.posedge
                    self.rx_vld.next = 0
                    for i in range(1,DEPTH):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=1<<i)
                        yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=0, tx_vld=1, stage_en=0)
                    yield self.clk.posedge
                    self.tx_rdy.next = 1
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=1, stage_en=0)
                    yield self.clk.posedge
                    self.tx_rdy.next = 0
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    yield self.clk.posedge

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.stop_rx = Signal(intbv(0)[dpt:])
                self.stop_tx = Signal(intbv(0)[dpt:])
                self.en = Signal(intbv(0)[dpt:])

                dut = getDut( pipeline_control,
                              rst = self.rst,
                              clk = self.clk,
                              rx_rdy = self.rx_rdy,
                              rx_vld = self.rx_vld,
                              tx_rdy = self.tx_rdy,
                              tx_vld = self.tx_vld,
                              stage_enable = self.en,
                              stop_rx = None,
                              stop_tx = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testContinuousWrite(self):
        ''' PIPE_CTRL: Continuous data '''
        DEPTH = [1,2,3,7,8,9]

        def stim(DEPTH):
            @instance
            def _inst():
                yield self.reset()
                yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                yield self.clk.posedge

                for _ in range(3):
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    self.rx_vld.next = 1
                    for i in range(1,DEPTH+1):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=2**i-1)
                        yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=0, tx_vld=1, stage_en=0)
                    self.rx_vld.next = 0
                    yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=0, tx_vld=1, stage_en=0)
                    self.tx_rdy.next = 1
                    for i in range(1,DEPTH+1):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=1, stage_en=(2**DEPTH-1)-(2**i-1))
                        yield self.clk.posedge
                    self.tx_rdy.next = 0
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    yield self.clk.posedge

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.stop_rx = Signal(intbv(0)[dpt:])
                self.stop_tx = Signal(intbv(0)[dpt:])
                self.en = Signal(intbv(0)[dpt:])

                dut = getDut( pipeline_control,
                              rst = self.rst,
                              clk = self.clk,
                              rx_rdy = self.rx_rdy,
                              rx_vld = self.rx_vld,
                              tx_rdy = self.tx_rdy,
                              tx_vld = self.tx_vld,
                              stage_enable = self.en,
                              stop_rx = None,
                              stop_tx = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testStopRX(self):
        ''' PIPE_CTRL: Stop RX '''
        DEPTH = [1,2,3,7,8,9]

        def stim(DEPTH):
            @instance
            def _inst():
                yield self.reset()
                yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                yield self.clk.posedge

                yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)

                for i in range(DEPTH):
                    # Stage i does not need input data, runs without waiting for stage i-1, and generates data to stage i+1
                    self.stop_rx.next = 1<<i
                    for j in range(i,DEPTH):
                        yield self.pipe_state_check(rx_rdy=0, tx_vld=0, stage_en=(2**(j+1)-1)-(2**i-1))
                        yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=0, tx_vld=1, stage_en=0)
                    self.stop_rx.next = 0
                    yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=0, tx_vld=1, stage_en=0)
                    self.tx_rdy.next = 1
                    # Drain the pipeline
                    for j in range(i,DEPTH):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=1, stage_en=(2**DEPTH-1)-(2**(j+1)-1))
                        yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    self.tx_rdy.next = 0
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    yield self.clk.posedge

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.stop_rx = Signal(intbv(0)[dpt:])
                self.stop_tx = Signal(intbv(0)[dpt:])
                self.en = Signal(intbv(0)[dpt:])

                dut = getDut( pipeline_control,
                              rst = self.rst,
                              clk = self.clk,
                              rx_rdy = self.rx_rdy,
                              rx_vld = self.rx_vld,
                              tx_rdy = self.tx_rdy,
                              tx_vld = self.tx_vld,
                              stage_enable = self.en,
                              stop_rx = self.stop_rx,
                              stop_tx = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testStopTX(self):
        ''' PIPE_CTRL: Stop TX '''
        DEPTH = [1,2,3,7,8,9]

        def stim(DEPTH):
            @instance
            def _inst():
                yield self.reset()
                yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                yield self.clk.posedge

                yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)

                for i in range(DEPTH):
                    self.stop_tx.next = 1<<i
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    self.rx_vld.next = 1
                    for j in range(i+1):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=2**(j+1)-1)
                        yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=2**(i+1)-1)
                    yield self.clk.posedge
                    self.rx_vld.next = 0
                    for j in range(i):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=(2**(i+1)-1)-(2**(j+1)-1))
                        yield self.clk.posedge
                    self.stop_tx.next = 0
                    for j in range(i+1,DEPTH):
                        yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=1<<j)
                        yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=0, tx_vld=1, stage_en=0)
                    self.tx_rdy.next = 1
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=1, stage_en=0)
                    yield self.clk.posedge
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)
                    self.tx_rdy.next = 0
                    yield self.pipe_state_check(rx_rdy=1, tx_vld=0, stage_en=0)

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.stop_rx = Signal(intbv(0)[dpt:])
                self.stop_tx = Signal(intbv(0)[dpt:])
                self.en = Signal(intbv(0)[dpt:])

                dut = getDut( pipeline_control,
                              rst = self.rst,
                              clk = self.clk,
                              rx_rdy = self.rx_rdy,
                              rx_vld = self.rx_vld,
                              tx_rdy = self.tx_rdy,
                              tx_vld = self.tx_vld,
                              stage_enable = self.en,
                              stop_rx = None,
                              stop_tx = self.stop_tx)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testPipelineControl']
    unittest.main()


from myhdl import *

def rom(addr, dout, CONTENT):
    ''' CONTENT == tuple of non-sparse values '''
    @always_comb
    def read():
        dout.next = CONTENT[int(addr)]

    return read


def ram_sp_rf(clk, we, addr, di, do):
    ''' RAM: Single-Port, Read-First '''

    memL = [Signal(intbv(0)[len(di):]) for _ in range(2**len(addr))]

    @always(clk.posedge)
    def write():
        if we:
            memL[int(addr)].next = di

        do.next = memL[int(addr)]

    return write


def ram_sp_wf(clk, we, addr, di, do):
    ''' RAM: Single-Port, Write-First '''

    memL = [Signal(intbv(0)[len(di):]) for _ in range(2**len(addr))]

    @always(clk.posedge)
    def write():        
        if we:
            memL[int(addr)].next = di
            do.next              = di
        else:
            do.next = memL[int(addr)]

    return write


def ram_sp_ar(clk, we, addr, di, do):
    ''' RAM: Single-Port, Asynchronous Read '''

    memL = [Signal(intbv(0)[len(di):]) for _ in range(2**len(addr))]

    @always(clk.posedge)
    def write():
        if we:
            memL[int(addr)].next = di

    @always_comb
    def read():
        do.next = memL[int(addr)]

    return write, read


def ram_sdp_rf(clk, we, addrw, addrr, di, do):
    ''' RAM: Simple-Dual-Port, Read-First '''

    memL = [Signal(intbv(0)[len(di):]) for _ in range(2**len(addrr))]

    @always(clk.posedge)
    def write():
        if we:
            memL[int(addrw)].next = di
        do.next = memL[int(addrr)]

    return write


def ram_sdp_wf(clk, we, addrw, addrr, di, do):
    ''' RAM: Simple-Dual-Port, Write-First '''

    memL = [Signal(intbv(0)[len(di):]) for _ in range(2**len(addrr))]
    addrr_r = Signal(intbv(0)[len(addrr):])

    @always(clk.posedge)
    def write():
        if we:
            memL[int(addrw)].next = di
        addrr_r.next          = addrr

    @always_comb
    def read():
        do.next = memL[int(addrr_r)]

    return write, read


def ram_sdp_ar(clk, we, addrw, addrr, di, do):
    ''' RAM: Simple-Dual-Port, Asynchronous Read'''

    memL = [Signal(intbv(0)[len(di):]) for _ in range(2**len(addrr))]

    @always(clk.posedge)
    def write():
        if we:
            memL[int(addrw)].next = di

    @always_comb
    def read():
        do.next = memL[int(addrr)]

    return write, read


def ram_dp_rf(clka, clkb, wea, web, addra, addrb, dia, dib, doa, dob):
    ''' RAM: Dual-Port, Read-First '''

    memL = [Signal(intbv(0)[len(dia):]) for _ in range(2**len(addra))]

    @always(clka.posedge)
    def writea():
        if wea:
            memL[int(addra)].next = dia

        doa.next = memL[int(addra)]

    @always(clkb.posedge)
    def writeb():
        if web:
            memL[int(addrb)].next = dib

        dob.next = memL[int(addrb)]

    return writea, writeb


def ram_dp_wf(clka, clkb, wea, web, addra, addrb, dia, dib, doa, dob):
    ''' RAM: Dual-Port, Write-First '''

    memL = [Signal(intbv(0)[len(dia):]) for _ in range(2**len(addra))]

    @always(clka.posedge)
    def writea():
        if wea:
            memL[int(addra)].next = dia
            doa.next              = dia
        else:
            doa.next = memL[int(addra)]

    @always(clkb.posedge)
    def writeb():
        if web:
            memL[int(addrb)].next = dib
            dob.next              = dib
        else:
            dob.next = memL[int(addrb)]

    return writea, writeb


def ram_dp_ar(clka, clkb, wea, web, addra, addrb, dia, dib, doa, dob):
    ''' RAM: Dual-Port, Asynchronous Read '''

    memL = [Signal(intbv(0)[len(dia):]) for _ in range(2**len(addra))]

    @always(clka.posedge)
    def writea():
        if wea:
            memL[int(addra)].next = dia

    @always_comb
    def reada():
        doa.next = memL[int(addra)]

    @always(clkb.posedge)
    def writeb():
        if web:
            memL[int(addrb)].next = dib

    @always_comb
    def readb():
        dob.next = memL[int(addrb)]

    return writea, reada, writeb, readb


#===============================================================================
# Conversion functions
#===============================================================================

def convert_rom(ADDR_WIDTH=8, DATA_WIDTH=8, CONTENT=(4,5,6,7)):
    ''' Convert ROM'''
    addr = Signal(intbv(0)[ADDR_WIDTH:])
    dout = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(rom, addr, dout, CONTENT)


def convert_ram_sp_rf(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Single-Port, Read-First '''
    clk = Signal(bool(0))
    we = Signal(bool(0))
    addr = Signal(intbv(0)[ADDR_WIDTH:])
    di = Signal(intbv(0)[DATA_WIDTH:])
    do = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_sp_rf, clk, we, addr, di, do)


def convert_ram_sp_wf(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Single-Port, Write-First '''
    clk = Signal(bool(0))
    we = Signal(bool(0))
    addr = Signal(intbv(0)[ADDR_WIDTH:])
    di = Signal(intbv(0)[DATA_WIDTH:])
    do = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_sp_wf, clk, we, addr, di, do)


def convert_ram_sp_ar(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Single-Port, Asynchronous Read '''
    clk = Signal(bool(0))
    we = Signal(bool(0))
    addr = Signal(intbv(0)[ADDR_WIDTH:])
    di = Signal(intbv(0)[DATA_WIDTH:])
    do = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_sp_ar, clk, we, addr, di, do)


def convert_ram_sdp_rf(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Simple-Dual-Port, Read-First '''
    clk = Signal(bool(0))
    we = Signal(bool(0))
    addrw = Signal(intbv(0)[ADDR_WIDTH:])
    addrr = Signal(intbv(0)[ADDR_WIDTH:])
    di = Signal(intbv(0)[DATA_WIDTH:])
    do = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_sdp_rf, clk, we, addrw, addrr, di, do)


def convert_ram_sdp_wf(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Simple-Dual-Port, Write-First '''
    clk = Signal(bool(0))
    we = Signal(bool(0))
    addrw = Signal(intbv(0)[ADDR_WIDTH:])
    addrr = Signal(intbv(0)[ADDR_WIDTH:])
    di = Signal(intbv(0)[DATA_WIDTH:])
    do = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_sdp_wf, clk, we, addrw, addrr, di, do)


def convert_ram_sdp_ar(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Simple-Dual-Port, Asynchronous Read'''
    clk = Signal(bool(0))
    we = Signal(bool(0))
    addrw = Signal(intbv(0)[ADDR_WIDTH:])
    addrr = Signal(intbv(0)[ADDR_WIDTH:])
    di = Signal(intbv(0)[DATA_WIDTH:])
    do = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_sdp_ar, clk, we, addrw, addrr, di, do)


def convert_ram_dp_rf(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Dual-Port, Read-First '''
    clka = Signal(bool(0))
    clkb = Signal(bool(0))
    wea = Signal(bool(0))
    web = Signal(bool(0))
    addra = Signal(intbv(0)[ADDR_WIDTH:])
    addrb = Signal(intbv(0)[ADDR_WIDTH:])
    dia = Signal(intbv(0)[DATA_WIDTH:])
    dib = Signal(intbv(0)[DATA_WIDTH:])
    doa = Signal(intbv(0)[DATA_WIDTH:])
    dob = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_dp_rf, clka, clkb, wea, web, addra, addrb, dia, dib, doa, dob)


def convert_ram_dp_wf(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Dual-Port, Write-First '''
    clka = Signal(bool(0))
    clkb = Signal(bool(0))
    wea = Signal(bool(0))
    web = Signal(bool(0))
    addra = Signal(intbv(0)[ADDR_WIDTH:])
    addrb = Signal(intbv(0)[ADDR_WIDTH:])
    dia = Signal(intbv(0)[DATA_WIDTH:])
    dib = Signal(intbv(0)[DATA_WIDTH:])
    doa = Signal(intbv(0)[DATA_WIDTH:])
    dob = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_dp_wf, clka, clkb, wea, web, addra, addrb, dia, dib, doa, dob)


def convert_ram_dp_ar(ADDR_WIDTH=8, DATA_WIDTH=8):
    ''' Convert RAM: Dual-Port, Asynchronous Read '''
    clka = Signal(bool(0))
    clkb = Signal(bool(0))
    wea = Signal(bool(0))
    web = Signal(bool(0))
    addra = Signal(intbv(0)[ADDR_WIDTH:])
    addrb = Signal(intbv(0)[ADDR_WIDTH:])
    dia = Signal(intbv(0)[DATA_WIDTH:])
    dib = Signal(intbv(0)[DATA_WIDTH:])
    doa = Signal(intbv(0)[DATA_WIDTH:])
    dob = Signal(intbv(0)[DATA_WIDTH:])
    toVerilog(ram_dp_ar, clka, clkb, wea, web, addra, addrb, dia, dib, doa, dob)




if __name__ == "__main__":
    convert_rom()
    convert_ram_sp_rf()
    convert_ram_sp_wf()
    convert_ram_sp_ar()
    convert_ram_sdp_rf()
    convert_ram_sdp_wf()
    convert_ram_sdp_ar()
    convert_ram_dp_rf()
    convert_ram_dp_wf()
    convert_ram_dp_ar()




from myhdl import ResetSignal, delay

# This code is a shameless copy of Christopher Felton's _reset class: https://github.com/cfelton/minnesota/blob/master/mn/system/_reset.py
class ResetSync(ResetSignal):
    ''' Synchronous reset '''
    def __init__(self, clk, val, active):
        self.clk = clk
        ResetSignal.__init__(self, val, active, async=False)

    def pulse(self, length_cc=10):
        self.next = self.active
        for _ in range(length_cc):
            yield self.clk.posedge
        self.next = not self.active

    def set(self):
        self.next = self.active

    def clear(self):
        self.next = not self.active
